from unittest.mock import call, patch

from arkindex.documents.models import Element
from arkindex.documents.tasks import add_parent_selection
from arkindex.project.tests import FixtureTestCase
from arkindex.users.models import User


class TestMoveSelection(FixtureTestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.user2 = User.objects.create(email='user2@test.test', display_name='User 2', verified_email=True)
        cls.parent1 = cls.corpus.elements.get(name='Volume 1')
        cls.parent2 = cls.corpus.elements.get(name='Volume 2')
        cls.page1 = cls.corpus.elements.get(name='Volume 1, page 1r')
        cls.page2 = cls.corpus.elements.get(name='Volume 1, page 2r')
        cls.page3 = cls.corpus.elements.get(name='Volume 1, page 1v')

    def test_no_rq_job(self):
        with self.assertRaises(AssertionError) as ctx:
            add_parent_selection(corpus_id=self.corpus.id, parent=self.parent2)
        self.assertEqual(str(ctx.exception), 'This task can only be run in a RQ job context.')

    @patch('arkindex.documents.tasks.get_current_job')
    def test_no_user_id(self, job_mock):
        job_mock.return_value.user_id = None
        with self.assertRaises(AssertionError) as ctx:
            add_parent_selection(corpus_id=self.corpus.id, parent=self.parent2)
        self.assertEqual(str(ctx.exception), 'This task requires a user ID to be defined on the RQ job.')

    @patch('arkindex.documents.tasks.get_current_job')
    def test_add_parent_selected_elements(self, job_mock):
        job_mock.return_value.user_id = self.user.id

        self.user.selected_elements.set([self.page1, self.page2])
        # Another user's selection should not be used
        self.user2.selected_elements.set([self.page3])

        add_parent_selection(corpus_id=self.corpus.id, parent=self.parent2)

        self.assertEqual(job_mock.call_count, 1)
        self.assertEqual(job_mock().set_progress.call_count, 2)
        self.assertListEqual(job_mock().set_progress.call_args_list, [
            call(.0),
            call(.5),
        ])

        self.assertQuerysetEqual(
            Element.objects.get_ascending(self.page1.id),
            [self.parent1, self.parent2],
            ordered=False,
        )
        self.assertQuerysetEqual(
            Element.objects.get_ascending(self.page2.id),
            [self.parent1, self.parent2],
            ordered=False,
        )
        self.assertQuerysetEqual(
            Element.objects.get_ascending(self.page3.id),
            [self.parent1],
            ordered=False,
        )
