import uuid
from datetime import datetime, timedelta, timezone

import sqlparse
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, Element, ElementType, EntityType, MetaData, MetaType, Transcription
from arkindex.process.models import WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import User


class TestListElements(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.worker_version = WorkerVersion.objects.get(worker__slug='reco')

    def _get_prefetch_data(self, query):
        """
        Retrieve a set of UUIDs or None, and the string holding them in the query,
        from an SQL query string generated by Django's prefetching.
        This only handles models that have UUIDs as primary keys.

        This allows you to reuse it in assertExactQueries:

        with self.assertExactQueries('file.sql', params={…}) as ctx:
            expected_ids = set(...)
            query = next(query['sql'] for query in ctx if 'images_image' in query['sql'])
            actual_ids = self._get_prefetch_data(query)
            self.assertSetEqual(expected_ids, actual_ids)
        """
        # sqlparse.parse returns a tuple of statements, we will only have one
        parsed_query, = sqlparse.parse(query)

        # Get the where clause
        where = next(token for token in parsed_query.tokens if isinstance(token, sqlparse.sql.Where))
        # Find the parentheses after `IN`
        parens = next(token for token in where.tokens if isinstance(token, sqlparse.sql.Parenthesis))
        # Find the identifier list
        id_list = next(token for token in parens.tokens if isinstance(token, sqlparse.sql.IdentifierList))

        # Parse the identifier list, handling both UUIDs and NULL.
        ids = set()
        for id_token in id_list.tokens:
            # In our case, the Identifier is a string parsed as a UUID: it is made of three Tokens.
            # The first token is the UUID as a string, the second is the cast operator (::) and the third is the UUID type.
            # We grab the string, remove its quotes, and parse it as a Python UUID.
            if isinstance(id_token, sqlparse.sql.Identifier):
                ids.add(uuid.UUID(id_token.tokens[0].value.strip("'")))
            # NULL is parsed as a single Token class with a Keyword ttype.
            elif id_token.match(sqlparse.tokens.Keyword, "NULL"):
                ids.add(None)

        return ids

    def test_list_elements_sql(self):
        """
        Test ListElements with assertExactQueries.

        ListElements prefetches images, servers and element types in two separate queries.
        Django gathers image and type IDs from elements in an unordered manner, and performs the queries in a random
        order, which means that we cannot easily put the prefetch queries in the SQL fixture.

        To overcome this, this test uses sqlparse to look through the AST for the prefetch query and find the list of
        images and types, then builds a list of UUIDs and compares it against an expected list of IDs.
        It then removes the queries from the assertExactQueries context so that they are not included in the
        SQL fixture assertion.

        Future maintainers: If you want to dig into a prefetch query, the simplest way is to explore the result of
        sqlparse.parse in an interactive shell.  You could copy-paste an SQL string from the file in sql_validation,
        or run pdb straight from this test:

        image_query = sql_parse.parse(…)
        import pdb
        pdb.set_trace()

        Then use `p image_query` to see the query, `p image_query.tokens` to see its child tokens, etc.
        You can access an image ID directly using `p image_query.tokens[-1].tokens[-1].tokens[-1].tokens[1].tokens[0].value`.
        Learn more about sqlparse tokens here: https://sqlparse.readthedocs.io/en/latest/analyzing/
        """

        expected_elements = self.corpus.elements.all().order_by('name', 'id')[:20]
        # Since Django 3.2, NULL values are not prefetched since a primary key cannot be NULL.
        expected_image_ids = set(filter(None, expected_elements.values_list('image_id', flat=True)))
        expected_type_ids = set(expected_elements.values_list('type_id', flat=True))

        with self.assertExactQueries('list_elements.sql', params={'corpus_id': self.corpus.id}) as ctx:
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

            # Test the image prefetch separately to avoid the aforementioned issues
            image_index, image_query = next(
                (i, query['sql'])
                for i, query in enumerate(ctx)
                if 'images_image' in query['sql']
            )
            actual_image_ids = self._get_prefetch_data(image_query)
            self.assertSetEqual(expected_image_ids, actual_image_ids, 'Unexpected prefetched image IDs')

            # Django's prefetch queries can be in a random order; we just drop them from the captured queries as we
            # asserted on them manually, so that they can be removed from the SQL fixture entirely.
            # `ctx[i]` just forwards to `ctx.captured_queries[i]`, which is a property using `ctx.connection.queries`.
            # `queries` is a property on a Django connection that is defined as `list(self.queries_log)`, and
            # `queries_log` is a collections.deque instance.
            # To drop the queries, we need to delete from that deque directly.
            # As assertExactQueries might have been triggered after some queries were already logged by the Django
            # connection, it has an 'initial_queries' index which is used in .captured_queries, so we have to make
            # our indexes relative to that.
            del ctx.connection.queries_log[ctx.initial_queries + image_index]

            # Rinse and repeat for the type prefetch.
            type_index, type_query = next(
                (i, query['sql'])
                for i, query in enumerate(ctx)
                if 'documents_elementtype' in query['sql']
            )
            actual_type_ids = self._get_prefetch_data(type_query)
            self.assertSetEqual(expected_type_ids, actual_type_ids, 'Unexpected prefetched type IDs')
            del ctx.connection.queries_log[ctx.initial_queries + type_index]

    def test_list_elements_filter_name(self):
        nameSelected = 'ol'
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'name': nameSelected}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        for result in results:
            self.assertIn(nameSelected, result['name'])

    def test_list_elements_filter_creator_email(self):
        page_type = ElementType.objects.get(corpus=self.corpus, slug='page')
        user2 = User.objects.get(email='user2@user.fr')
        self.corpus.elements.create(type=page_type, name='Test Element', creator=user2)
        mail = 'user2@user.fr'
        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'creator_email': mail}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]['name'], 'Test Element')

    def test_list_elements_filter_creator_email_case_insensitive(self):
        page_type = ElementType.objects.get(corpus=self.corpus, slug='page')
        user2 = User.objects.get(email='user2@user.fr')
        self.corpus.elements.create(type=page_type, name='Test Element', creator=user2)
        mail = 'UsER2@user.fr'
        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'creator_email': mail}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        self.assertEqual(len(results), 1)
        self.assertEqual(results[0]['name'], 'Test Element')

    def test_list_elements_filter_creator_email_does_not_exist(self):
        """
        If there is no user with the given email, there is no error but no results are returned
        """
        mail = 'ohno@user.fr'
        with self.assertNumQueries(2):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'creator_email': mail}
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        self.assertEqual(len(results), 0)

    def test_list_elements_corpus_uuid4(self):
        """
        Ensure the corpus filter allows any UUID, not just version 4 UUIDs.
        Using the version=x argument on uuid.UUID causes the version and variant bits
        of a UUID to be changed on instantiation; since we generate v4 RFC 4122-compliant UUIDs,
        this would work in most cases - but we could have any UUID version or variant,
        generated by some other tool, which would cause strange 404 errors.
        """
        str_id = '12345678-1234-5678-1234-567812345678'
        self.assertNotEqual(
            uuid.UUID(str_id),
            uuid.UUID(str_id, version=4),
        )
        corpus = Corpus.objects.create(id=uuid.UUID(str_id), name='A corpus', public=True)
        element_type = corpus.types.create(slug='element', display_name='Generic element')
        element = corpus.elements.create(type=element_type, name='Something')

        response = self.client.get(reverse('api:corpus-elements', kwargs={'corpus': str_id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(element.id))
        self.assertEqual(data['results'][0]['corpus']['id'], str_id)

    def test_list_elements_filter_type(self):
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'type': 'act'}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        for result in results:
            self.assertEqual(result['type'], 'act')

    def test_list_with_has_children(self):
        # Iterate over every page of results of this endpoint
        page = 1
        results = {}
        while True:
            with self.assertNumQueries(7):
                response = self.client.get(
                    reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                    data={'with_has_children': True, 'page': page},
                )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
            data = response.json()
            for element in data['results']:
                results[element['name']] = element['has_children']
            if not data['next']:
                break
            page += 1

        self.assertDictEqual(
            results,
            {
                'Act 1': True,
                'Act 2': True,
                'Act 3': True,
                'Act 4': True,
                'Act 5': True,
                'Text line': False,
                'Surface A': False,
                'Surface B': False,
                'Surface C': False,
                'Surface D': False,
                'Surface E': False,
                'Surface F': False,
                'Volume 1': True,
                'Volume 1, page 1r': True,
                'Volume 1, page 1v': True,
                'Volume 1, page 2r': True,
                'Volume 2': True,
                'Volume 2, page 1r': False,
                'Volume 2, page 1v': False,
                'Volume 2, page 2r': False,
                'DATUM': False,
                'PARIS': False,
                'ROY': False,
            }
        )

    def test_list_top_level(self):
        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'top_level': True},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            ['Volume 1', 'Volume 2'],
        )

    def test_list_top_level_default_type(self):
        """
        Do not use the corpus top level type to filter top level elements when available;
        the frontend or API users are in charge of that.
        """
        # Add a top level element of type page
        element_type = self.corpus.types.create(slug='top_level', display_name='Some type')
        self.corpus.elements.create(type=element_type, name='Generic element')
        self.corpus.top_level_type = element_type
        self.corpus.save()

        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'top_level': True},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            ['Generic element', 'Volume 1', 'Volume 2'],
        )

    def test_list_top_level_non_folder(self):
        """
        The top_level filter should allow non-folder elements
        """
        self.corpus.elements.create(type=self.corpus.types.filter(folder=False).first(), name='lonely page')

        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'top_level': True},
            )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            ['lonely page', 'Volume 1', 'Volume 2']
        )

    def test_list_non_top_level(self):
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'top_level': False},
            )

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            [
                'Act 1',
                'Act 2',
                'Act 3',
                'Act 4',
                'Act 5',
                'DATUM',
                'DATUM',
                'DATUM',
                'PARIS',
                'PARIS',
                'PARIS',
                'ROY',
                'ROY',
                'ROY',
                'Surface A',
                'Surface B',
                'Surface C',
                'Surface D',
                'Surface E',
                'Surface F',
            ]
        )

    def test_list_elements_filter_no_worker_version(self):
        self.corpus.elements.update(worker_version=self.worker_version)
        self.corpus.elements.filter(name__contains='Volume 2').update(worker_version=None)

        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'worker_version': False},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            ['Volume 2', 'Volume 2, page 1r', 'Volume 2, page 1v', 'Volume 2, page 2r'],
        )

        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'worker_version': 0},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            ['Volume 2', 'Volume 2, page 1r', 'Volume 2, page 1v', 'Volume 2, page 2r'],
        )

    def test_list_elements_filter_worker_version(self):
        self.corpus.elements.filter(name__contains='Volume 2').update(worker_version=self.worker_version)

        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'worker_version': str(self.worker_version.id)},
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            ['Volume 2', 'Volume 2, page 1r', 'Volume 2, page 1v', 'Volume 2, page 2r'],
        )

    def test_list_elements_filter_worker_version_validation(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'worker_version': 'blah'},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version': ['Invalid UUID']})

        with self.assertNumQueries(2):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'worker_version': uuid.uuid4()},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'worker_version': ['This worker version does not exist.']})

    def test_list_elements_filter_modified_since_not_modified(self):
        with self.assertNumQueries(2):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                HTTP_IF_MODIFIED_SINCE='Thu, 02 Apr 2099 13:37:42 GMT',
            )
        self.assertEqual(response.status_code, status.HTTP_304_NOT_MODIFIED)

    def test_list_elements_filter_modified_since_bad_format(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                HTTP_IF_MODIFIED_SINCE='AAAAAAAAAAAAAAAAA',
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'If-Modified-Since': ['Bad date format']})

    def test_list_elements_filter_modified_since(self):
        self.corpus.elements.filter(name='Volume 1').update(updated='2099-04-02T13:37:43Z')
        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                HTTP_IF_MODIFIED_SINCE='Thu, 02 Apr 2099 13:37:42 GMT',
            )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertEqual(response.json()['count'], self.corpus.elements.count())

    def test_list_elements_filter_metadata_name(self):
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'metadata_name': 'folio'}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            [
                'Volume 1, page 1r',
                'Volume 1, page 1v',
                'Volume 1, page 2r',
                'Volume 2, page 1r',
                'Volume 2, page 1v',
                'Volume 2, page 2r',
            ]
        )

    def test_list_elements_filter_metadata_value(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'metadata_value': '1v'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'metadata_value': ['This filter is not supported without metadata_name.'],
        })

    def test_list_elements_filter_metadata_name_value(self):
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'metadata_name': 'folio', 'metadata_value': '1v'}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            ['Volume 1, page 1v', 'Volume 2, page 1v']
        )

    def test_list_elements_filter_metadata_operator(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'metadata_operator': 'gt'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'metadata_operator': ['This option is not supported without metadata_value.']
        })

    def test_list_elements_filter_metadata_operator_value(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'metadata_operator': 'gt', 'metadata_value': 4}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'metadata_value': ['This filter is not supported without metadata_name.'],
        })

    def test_list_elements_filter_metadata_operator_invalid(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'metadata_operator': 'aaaaaa'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'metadata_operator': [
                'This metadata operator is not supported.',
                'This option is not supported without metadata_value.',
            ]
        })

    def test_list_elements_filter_numeric_metadata_invalid_value(self):
        with self.assertNumQueries(1):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={'metadata_name': 'a', 'metadata_value': 'lol', 'metadata_operator': 'gt'}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'metadata_value': ['Value must be a float when using a numeric comparison operator.'],
        })

    def test_list_elements_filter_numeric_metadata(self):
        element_type = self.corpus.types.first()
        for i in range(1, 6):
            element = self.corpus.elements.create(type=element_type, name=f'Element {i}')
            element.metadatas.create(name='number', type=MetaType.Numeric, value=str(i))
            element.metadatas.create(name='number', type=MetaType.Text, value='a')

        # operator, value, expected element names
        parameters = [
            ('lt', 2, ['Element 1']),
            ('lte', 2, ['Element 1', 'Element 2']),
            ('gt', 3, ['Element 4', 'Element 5']),
            ('gte', 3, ['Element 3', 'Element 4', 'Element 5']),
        ]

        for operator, value, expected_elements in parameters:
            with self.subTest(operator=operator, value=value):
                with self.assertNumQueries(4):
                    response = self.client.get(
                        reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                        data={
                            'metadata_name': 'number',
                            'metadata_value': value,
                            'metadata_operator': operator,
                        }
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element['name'] for element in response.json()['results']],
                    expected_elements
                )

    def test_list_elements_filter_metadata_contains(self):
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                data={
                    'metadata_name': 'folio',
                    'metadata_operator': 'contains',
                    'metadata_value': 'v',
                }
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertListEqual(
            [element['name'] for element in response.json()['results']],
            ['Volume 1, page 1v', 'Volume 2, page 1v']
        )

    def test_list_elements_filter_rotation_angle(self):
        element_type = self.corpus.types.first()
        self.corpus.elements.create(type=element_type, name='not rotated')
        self.corpus.elements.create(type=element_type, name='rotated', rotation_angle=180)

        cases = [
            (0, ['not rotated']),
            (1, []),
            (180, ['rotated']),
            (181, [])
        ]
        for rotation_angle, expected_elements in cases:
            with self.subTest(rotation_angle=rotation_angle):
                # Requests that return no elements only make 2 SQL queries
                num_queries = 4 if len(expected_elements) else 2
                with self.assertNumQueries(num_queries):
                    response = self.client.get(
                        reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                        data={
                            'name': 'rotated',
                            'rotation_angle': rotation_angle,
                        }
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element['name'] for element in response.json()['results']],
                    expected_elements
                )

    def test_list_elements_filter_rotation_angle_invalid(self):
        cases = [
            ('to the left', "invalid literal for int() with base 10: 'to the left'"),
            (-1, 'A rotation angle must be between 0 and 359 degrees'),
            (360, 'A rotation angle must be between 0 and 359 degrees'),
            (54.2, "invalid literal for int() with base 10: '54.2'"),
            (4.14j, "invalid literal for int() with base 10: '4.14j'"),
        ]
        for rotation_angle, expected_error in cases:
            with self.subTest(rotation_angle=rotation_angle):
                with self.assertNumQueries(1):
                    response = self.client.get(
                        reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                        data={
                            'rotation_angle': rotation_angle,
                        }
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertEqual(response.json(), {
                    'rotation_angle': [expected_error]
                })

    def test_list_elements_filter_mirrored(self):
        element_type = self.corpus.types.first()
        self.corpus.elements.create(type=element_type, name='mirrored', mirrored=True)
        self.corpus.elements.create(type=element_type, name='not mirrored', mirrored=False)

        cases = [
            (False, ['not mirrored']),
            (True, ['mirrored']),
        ]
        for mirrored, expected_elements in cases:
            with self.subTest(mirrored=mirrored):
                with self.assertNumQueries(4):
                    response = self.client.get(
                        reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                        data={
                            'name': 'mirrored',
                            'mirrored': mirrored,
                        }
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element['name'] for element in response.json()['results']],
                    expected_elements
                )

    def test_list_elements_with_corpus_false(self):
        with self.assertNumQueries(5):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id})
                + '?with_corpus=false'
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertSetEqual(
            set(element.get('corpus') for element in response.json()['results']),
            {None}
        )

    def test_list_elements_page_size(self):
        """
        Corpus elements can be listed with a large page size
        """
        corpus = Corpus.objects.create(name="A large corpus", public=True)
        folder_type = corpus.types.create(slug='folder', display_name='Folder')
        # Create 99 folder on this corpus
        Element.objects.bulk_create(
            Element(corpus=corpus, name=f"folder {i}", type=folder_type)
            for i in range(1, 100)
        )
        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': corpus.id}),
                {'page_size': 50}
            )
        data = response.json()
        self.assertEqual(len(data['results']), 50)
        self.assertEqual(data['count'], 99)

    def test_list_elements_page_size_limit(self):
        """
        Page size is limited to 500 items
        """
        corpus = Corpus.objects.create(name="A large corpus", public=True)
        folder_type = corpus.types.create(slug='folder', display_name='Folder')
        # Create 99 folder on this corpus
        Element.objects.bulk_create(
            Element(corpus=corpus, name=f"folder {i}", type=folder_type)
            for i in range(1, 1000)
        )
        with self.assertNumQueries(4):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': corpus.id}),
                {'page_size': 800}
            )
        data = response.json()
        self.assertEqual(len(data['results']), 500)
        self.assertEqual(data['count'], 999)

    def test_list_elements_invalid_order(self):
        cases = [
            ({'order': 'blah', 'order_direction': 'asc'}, {'order': ['Unknown sorting field']}),
            ({'order': 'name', 'order_direction': 'left'}, {'order_direction': ['Unknown sorting direction']}),
            ({'order': 'blah', 'order_direction': 'left'}, {'order': ['Unknown sorting field']}),
        ]
        for params, expected in cases:
            with self.subTest(**params):
                response = self.client.get(
                    reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                    params,
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), expected)

    def test_list_elements_order_created(self):
        elements = list(self.corpus.elements.order_by('?').only('id'))
        created = datetime(2020, 2, 2, tzinfo=timezone.utc)
        for element in elements:
            element.created = created
            created += timedelta(seconds=1)
        Element.objects.bulk_update(elements, ['created'])

        cases = [
            ('asc', elements[:20]),
            ('desc', reversed(elements[-20:])),
        ]
        for direction, expected in cases:
            with self.subTest(direction=direction), self.assertNumQueries(5):
                response = self.client.get(
                    reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                    {'order': 'created', 'order_direction': direction}
                )
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element['id'] for element in response.json()['results']],
                    [str(element.id) for element in expected]
                )

    def test_list_elements_order_random(self):
        # Get all expected element IDs as strings
        element_ids = list(map(str, self.corpus.elements.filter(type__slug='page').values_list('id', flat=True)))

        # Repeat the API request multiple times; the element order should change at least once
        responses = set()
        for _ in range(5):
            with self.assertNumQueries(6):
                response = self.client.get(
                    reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                    {'type': 'page', 'order': 'random'},
                )
                self.assertEqual(response.status_code, status.HTTP_200_OK)

            data = response.json()

            # Checking for the expected element IDs will start failing randomly
            # if there is more than 1 page of results.
            self.assertIsNone(data['next'])
            actual_ids = [element['id'] for element in data['results']]
            self.assertCountEqual(element_ids, actual_ids)

            responses.add(tuple(actual_ids))

        # We know that all responses have the expected elements, so if the responses set
        # has more than one item, the ordering has changed between requests, so it is likely to be random.
        self.assertGreater(len(responses), 1)

    def test_list_elements_with_metadata(self):
        # Remove all metadata from a page; we should get an empty list
        page1r = self.corpus.elements.get(name='Volume 1, page 1r')
        page1r.metadatas.all().delete()
        # Add an entity to a metadata; it should be excluded
        page1v = self.corpus.elements.get(name='Volume 1, page 1v')
        meta = page1v.metadatas.get()
        location_type = EntityType.objects.get(name='location', corpus=self.corpus)
        meta.entity = self.corpus.entities.create(name='Something', type=location_type)
        meta.save()

        # Retrieve metadata UUIDs for assertions
        metadata_ids = dict(
            MetaData.objects
                    .filter(element__corpus=self.corpus, element__type__slug='page')
                    .values_list('element__name', 'id')
        )

        with self.assertNumQueries(7):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                {'with_metadata': True, 'type': 'page'},
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual({
            element['name']: element['metadata']
            for element in response.json()['results']
        }, {
            'Volume 1, page 1r': [],
            'Volume 1, page 1v': [],
            'Volume 1, page 2r': [
                {
                    'id': str(metadata_ids['Volume 1, page 2r']),
                    'name': 'folio',
                    'type': 'text',
                    'value': '2r',
                    'dates': [],
                }
            ],
            'Volume 2, page 1r': [
                {
                    'id': str(metadata_ids['Volume 2, page 1r']),
                    'name': 'folio',
                    'type': 'text',
                    'value': '1r',
                    'dates': [],
                }
            ],
            'Volume 2, page 1v': [
                {
                    'id': str(metadata_ids['Volume 2, page 1v']),
                    'name': 'folio',
                    'type': 'text',
                    'value': '1v',
                    'dates': [],
                }
            ],
            'Volume 2, page 2r': [
                {
                    'id': str(metadata_ids['Volume 2, page 2r']),
                    'name': 'folio',
                    'type': 'text',
                    'value': '2r',
                    'dates': [],
                }
            ],
        })

    def test_metadata_classification_filters(self):
        """
        Using classification filters should not cause metadata filters to be ignored
        """
        volume = self.corpus.elements.get(name='Volume 1')
        page = self.corpus.elements.get(name='Volume 1, page 1r')
        ml_class = self.corpus.ml_classes.create(name='Dallas')

        volume.classifications.create(ml_class=ml_class, confidence=1.0)
        page.classifications.create(ml_class=ml_class, confidence=0.5)

        # The volume and the page have the Dallas MLClass
        self.assertEqual(self.corpus.elements.filter(classifications__ml_class=ml_class).count(), 2)
        # All 6 pages have a folio metadata
        self.assertEqual(self.corpus.elements.filter(metadatas__name='folio').count(), 6)
        # Only the page has both the folio metadata and the Dallas MLClass
        self.assertEqual(self.corpus.elements.filter(classifications__ml_class=ml_class, metadatas__name='folio').count(), 1)

        with self.assertNumQueries(6):
            response = self.client.get(
                reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                {'metadata_name': 'folio', 'class_id': str(ml_class.id)}
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(data['count'], 1)
        self.assertEqual(data['results'][0]['id'], str(page.id))

    def test_transcription_confidence_invalid(self):
        cases = [
            ({'transcription_confidence': 'very'}, {'transcription_confidence': ["could not convert string to float: 'very'"]}),
            ({'transcription_confidence': 'nan'}, {'transcription_confidence': ['Confidence must be between 0 and 1']}),
            ({'transcription_confidence': 'inf'}, {'transcription_confidence': ['Confidence must be between 0 and 1']}),
            ({'transcription_confidence': '42'}, {'transcription_confidence': ['Confidence must be between 0 and 1']}),
            ({'transcription_confidence': '-.5'}, {'transcription_confidence': ['Confidence must be between 0 and 1']}),
            (
                {'transcription_confidence_operator': 'hah'},
                {'transcription_confidence_operator': [
                    'This operator is not supported.',
                    'This option is not supported without transcription_confidence.',
                ]}
            ),
            (
                {'transcription_confidence': '0.3', 'transcription_confidence_operator': 'lol'},
                {'transcription_confidence_operator': ['This operator is not supported.']}
            )
        ]
        for data, expected_errors in cases:
            with self.subTest(**data):
                response = self.client.get(reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}), data)
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), expected_errors)

    def test_transcription_confidence_filter(self):
        # All transcriptions in the fixtures have a confidence of 1. We update them to be able to test properly.
        self.assertEqual(Transcription.objects.all().count(), 10)
        for i, transcription in enumerate(Transcription.objects.order_by('text')):
            transcription.confidence = i / 10
            transcription.save()

        cases = [
            (
                {'transcription_confidence': 0.5},
                ['PARIS']
            ),
            (
                {'transcription_confidence': 0.51},
                []
            ),
            (
                {'transcription_confidence': 0.5, 'transcription_confidence_operator': 'gt'},
                ['PARIS', 'ROY', 'ROY', 'ROY']
            ),
            (
                {'transcription_confidence': 0.5, 'transcription_confidence_operator': 'lte'},
                ['DATUM', 'DATUM', 'DATUM', 'PARIS', 'PARIS', 'Volume 1, page 1r']
            )
        ]
        for data, element_names in cases:
            with self.subTest(**data):
                response = self.client.get(reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}), data)
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element['name'] for element in response.json()['results']],
                    element_names,
                )

    def test_transcription_worker_version_invalid(self):
        cases = [
            ('a', {'transcription_worker_version': ['Invalid UUID']}),
            (42, {'transcription_worker_version': ['Invalid UUID']}),
            (True, {'transcription_worker_version': ['Invalid UUID']}),
            ('aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa', {'transcription_worker_version': ['This worker version does not exist.']}),
        ]
        for transcription_worker_version, expected_errors in cases:
            with self.subTest(transcription_worker_version=transcription_worker_version):
                response = self.client.get(
                    reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                    {'transcription_worker_version': transcription_worker_version}
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), expected_errors)

    def test_transcription_worker_version_filter(self):
        dla = WorkerVersion.objects.get(worker__slug='dla')
        Transcription.objects.filter(text='PARIS').update(worker_version=None)
        Transcription.objects.filter(text='ROY').update(worker_version=dla)

        cases = [
            (str(self.worker_version.id), ['DATUM', 'DATUM', 'DATUM', 'Volume 1, page 1r']),
            (str(dla.id), ['ROY', 'ROY', 'ROY']),
            (False, ['PARIS', 'PARIS', 'PARIS']),
        ]
        for transcription_worker_version, element_names in cases:
            with self.subTest(transcription_worker_version=transcription_worker_version):
                response = self.client.get(
                    reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}),
                    {'transcription_worker_version': transcription_worker_version}
                )
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element['name'] for element in response.json()['results']],
                    element_names,
                )

    def test_confidence_invalid(self):
        cases = [
            ({'confidence': 'very'}, {'confidence': ["could not convert string to float: 'very'"]}),
            ({'confidence': 'nan'}, {'confidence': ['Confidence must be between 0 and 1']}),
            ({'confidence': 'inf'}, {'confidence': ['Confidence must be between 0 and 1']}),
            ({'confidence': '42'}, {'confidence': ['Confidence must be between 0 and 1']}),
            ({'confidence': '-.5'}, {'confidence': ['Confidence must be between 0 and 1']}),
            (
                {'confidence_operator': 'hah'},
                {'confidence_operator': [
                    'This operator is not supported.',
                    'This option is not supported without `confidence`.',
                ]}
            ),
            (
                {'confidence': '0.3', 'confidence_operator': 'lol'},
                {'confidence_operator': ['This operator is not supported.']}
            )
        ]
        for data, expected_errors in cases:
            with self.subTest(**data):
                response = self.client.get(reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}), data)
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), expected_errors)

    def test_confidence_filter(self):
        Element.objects.bulk_create([
            Element(
                corpus=self.corpus,
                type=self.corpus.types.get(slug='volume'),
                confidence=i / 10,
                name=f'Confidence {i / 10}'
            )
            for i in range(11)
        ])

        cases = [
            (
                {'confidence': 0.5},
                ['Confidence 0.5']
            ),
            (
                {'confidence': 0.51},
                []
            ),
            (
                {'confidence': 0.5, 'confidence_operator': 'gt'},
                ['Confidence 0.6', 'Confidence 0.7', 'Confidence 0.8', 'Confidence 0.9', 'Confidence 1.0']
            ),
            (
                {'confidence': 0.5, 'confidence_operator': 'lte'},
                ['Confidence 0.0', 'Confidence 0.1', 'Confidence 0.2', 'Confidence 0.3', 'Confidence 0.4', 'Confidence 0.5']
            )
        ]
        for data, element_names in cases:
            with self.subTest(**data):
                response = self.client.get(reverse('api:corpus-elements', kwargs={'corpus': self.corpus.id}), data)
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                self.assertListEqual(
                    [element['name'] for element in response.json()['results']],
                    element_names,
                )
