from collections import defaultdict
from textwrap import dedent

from django.db.models import Prefetch
from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from arkindex.process.models import (
    GitRef,
    ProcessMode,
    WorkerConfiguration,
    WorkerRun,
    WorkerVersion,
    WorkerVersionState,
)
from arkindex.process.serializers.imports import ProcessTrainingSerializer
from arkindex.process.serializers.workers import WorkerConfigurationSerializer, WorkerVersionSerializer
from arkindex.project.mixins import WorkerACLMixin
from arkindex.training.models import ModelVersion, ModelVersionState
from arkindex.training.serializers import ModelVersionLightSerializer

# To prevent each element worker to retrieve contextual information
# (process, worker version, model version…) with extra GET requests, we
# do serialize all the related information on WorkerRun serializers.


class WorkerRunSerializer(WorkerACLMixin, serializers.ModelSerializer):

    parents = serializers.ListField(
        child=serializers.UUIDField(),
        default=list,
        help_text='UUIDs of the WorkerRuns that this WorkerRun depends on. '
                  'The WorkerRuns must be on the same process, and the dependencies cannot create a cycle.',
    )

    worker_version = WorkerVersionSerializer(read_only=True, source='version')
    worker_version_id = serializers.PrimaryKeyRelatedField(
        write_only=True,
        source='version',
        queryset=(
            WorkerVersion
            .objects
            .select_related(
                # Required for the WorkerACLMixin
                'worker__repository',
                # Used by the WorkerVersionSerializer
                'worker__type',
                'revision__repo',
            )
            .prefetch_related(
                # Used by the WorkerVersionSerializer
                Prefetch(
                    'revision__refs',
                    queryset=GitRef.objects.select_related('repository')
                )
            )
        ),
        help_text='UUID of the WorkerVersion for this WorkerRun. '
                  'Only WorkerVersions in an `available` state may be set. '
                  'Contributor access to the worker is required.',
    )

    model_version = ModelVersionLightSerializer(read_only=True)
    model_version_id = serializers.PrimaryKeyRelatedField(
        queryset=ModelVersion.objects.none(),
        required=False,
        allow_null=True,
        write_only=True,
        source='model_version',
        style={'base_template': 'input.html'},
        help_text=dedent(
            """
            UUID of the ModelVersion for this WorkerRun, or `null` if none is set.

            Only ModelVersions in an `available` state may be set.

            Guest access to the model is required.
            For model versions without a `tag`, contributor access is required.
            """
        ),
    )

    configuration = WorkerConfigurationSerializer(read_only=True)
    configuration_id = serializers.PrimaryKeyRelatedField(
        queryset=WorkerConfiguration.objects.all(),
        required=False,
        allow_null=True,
        write_only=True,
        style={'base_template': 'input.html'},
        source='configuration',
        help_text='UUID of the WorkerConfiguration for this WorkerRun, or `null` if none is set. '
                  "Only a configuration of the WorkerVersion's worker may be set.",
    )

    process = ProcessTrainingSerializer(read_only=True)

    class Meta:
        model = WorkerRun
        fields = (
            'id',
            'parents',
            'worker_version_id',
            'worker_version',
            'process',
            'configuration_id',
            'configuration',
            'model_version_id',
            'model_version',
            'summary',
        )
        read_only_fields = (
            'id',
            'process',
            'worker_version',
            'configuration',
            'model_version',
            'summary',
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.context.get('request'):
            user = self.context['request'].user
            self.fields['model_version_id'].queryset = ModelVersion.objects.executable(user)

    def validate(self, data):
        data = super().validate(data)
        errors = defaultdict(list)
        worker_version, model_version, configuration = None, None, None

        if self.instance:
            process = self.instance.process
            worker_version = self.instance.version
            model_version = self.instance.model_version
            configuration = self.instance.configuration
        else:
            data['process'] = process = self.context['process']

        worker_version = data.get('version', worker_version)
        model_version = data.get('model_version', model_version)
        configuration = data.get('configuration', configuration)
        parents = data.get('parents', [])

        if process.mode not in (ProcessMode.Workers, ProcessMode.Dataset):
            errors['process_id'].append('WorkerRuns can only be created or updated on Workers or Dataset processes.')
        if process.has_tasks:
            errors['process_id'].append('WorkerRuns cannot be added or updated on processes that have already started.')

        if len(parents) != len(set(parents)):
            errors['parents'].append('The parents of a WorkerRun must be unique.')

        # The worker version is not editable on existing WorkerRun instances, so we only validate when creating a new one
        if not self.instance:
            # Set self.request to please the WorkerACLMixin
            self.request = self.context['request']
            if not self.has_execution_access(worker_version.worker):
                errors['worker_version_id'].append('You do not have an execution access to this worker.')
            elif worker_version.state != WorkerVersionState.Available:
                errors['worker_version_id'].append('This WorkerVersion is not in an Available state.')

        if configuration and configuration.worker_id != worker_version.worker.id:
            errors['configuration_id'].append('The configuration must be part of the same worker.')

        if model_version:
            if model_version.state != ModelVersionState.Available:
                errors['model_version_id'].append('This ModelVersion is not in an Available state.')

            if not worker_version.model_usage:
                errors['model_version_id'].append('This worker version does not support models.')

        existing_worker_run = process.worker_runs.filter(version=worker_version, model_version_id=model_version, configuration=configuration)
        if self.instance:
            # The current worker run cannot be a duplicate of itself!
            existing_worker_run = existing_worker_run.exclude(id=self.instance.id)
        if existing_worker_run.exists():
            errors['__all__'].append('A WorkerRun already exists on this process with the selected worker version, model version and configuration.')

        if errors:
            raise ValidationError(errors)

        # Since we only allow creating a WorkerRun on a process without any task, we know the last run is None,
        # without having to make any extra SQL queries to serialize it in the response.
        process.last_run = None

        return data


class WorkerRunEditSerializer(WorkerRunSerializer):

    class Meta(WorkerRunSerializer.Meta):
        # Same as WorkerRunSerializer, but the worker_version_id cannot be edited
        fields = tuple(set(WorkerRunSerializer.Meta.fields) - {'worker_version_id'})
