from unittest.mock import patch

from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.process.models import DataFile
from arkindex.project.tests import FixtureAPITestCase


@override_settings(AWS_STAGING_BUCKET='staging')
class TestFiles(FixtureAPITestCase):
    """
    Test file management
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.df = DataFile.objects.create(
            name='test.txt',
            size=42,
            content_type='text/plain',
            corpus=cls.corpus
        )

    def setUp(self):
        super().setUp()
        self.client.force_login(self.user)

    def test_file_list(self):
        response = self.client.get(reverse('api:file-list', kwargs={'pk': self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertIn('results', data)
        self.assertEqual(len(data['results']), 1)

        file = data['results'][0]
        self.assertIn('id', file)
        self.assertIn('name', file)
        self.assertIn('size', file)
        self.assertEqual(file['id'], str(self.df.id))
        self.assertEqual(file['name'], self.df.name)
        self.assertEqual(file['size'], self.df.size)

    def test_file_list_filter_out_trashed_files(self):
        DataFile.objects.create(
            name='test2.txt',
            size=42,
            content_type='text/plain',
            corpus=self.corpus,
            trashed=True
        )
        self.assertEqual(DataFile.objects.count(), 2)
        response = self.client.get(reverse('api:file-list', kwargs={'pk': self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertIn('results', data)
        self.assertEqual(len(data['results']), 1)

    def test_file_list_requires_login(self):
        self.client.logout()
        response = self.client.get(reverse('api:file-list', kwargs={'pk': self.corpus.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @patch('arkindex.project.aws.s3')
    def test_file_delete(self, s3_mock):
        self.assertTrue(DataFile.objects.exists())
        response = self.client.delete(reverse('api:file-retrieve', kwargs={'pk': self.df.id}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertTrue(DataFile.objects.exists())
        self.assertTrue(DataFile.objects.get().trashed)
        self.assertEqual(s3_mock.Object().delete.call_count, 0)

    def test_file_delete_requires_login(self):
        """
        Assert a file deletion requires a login
        """
        self.client.logout()
        self.assertEqual(DataFile.objects.count(), 1)
        response = self.client.delete(reverse('api:file-retrieve', kwargs={'pk': self.df.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(DataFile.objects.count(), 1)

    def test_file_delete_requires_verified(self):
        """
        Assert a file deletion requires being logged in with a verified email
        """
        self.user.verified_email = False
        self.user.save()
        self.assertEqual(DataFile.objects.count(), 1)
        response = self.client.delete(reverse('api:file-retrieve', kwargs={'pk': self.df.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertEqual(DataFile.objects.count(), 1)
