import boto3
from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Scope
from moto import mock_s3


class TestListBuckets(FixtureAPITestCase):

    def test_requires_login(self):
        with self.assertNumQueries(0):
            resp = self.client.get(reverse('api:bucket-list'))
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.user.user_scopes.create(scope=Scope.S3Ingest)
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            resp = self.client.get(reverse('api:bucket-list'))
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_requires_scope(self):
        self.assertFalse(self.user.user_scopes.exists())
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            resp = self.client.get(reverse('api:bucket-list'))
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    @override_settings(
        INGEST_S3_ACCESS_KEY=None,
        INGEST_S3_SECRET_KEY=None,
        INGEST_S3_ENDPOINT=None,
        INGEST_S3_REGION=None,
    )
    def test_missing_settings(self):
        self.user.user_scopes.create(scope=Scope.S3Ingest)
        self.client.force_login(self.user)
        with self.assertNumQueries(3), self.assertRaises(AssertionError):
            self.client.get(reverse('api:bucket-list'))

    @override_settings(
        INGEST_S3_ACCESS_KEY='just-let-me-in',
        INGEST_S3_SECRET_KEY='i-dunno',
        INGEST_S3_ENDPOINT=None,
        INGEST_S3_REGION='atlantis-1',
        INGEST_EXTRA_BUCKETS=[],
    )
    @mock_s3
    def test_list_buckets(self):
        self.user.user_scopes.create(scope=Scope.S3Ingest)
        self.client.force_login(self.user)

        # Create some buckets with moto to get some test data
        s3 = boto3.resource('s3', region_name='us-east-1')
        s3.create_bucket(Bucket='dank-memes')
        s3.create_bucket(Bucket='cursed-emojis')

        with self.assertNumQueries(3):
            resp = self.client.get(reverse('api:bucket-list'))
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.assertListEqual(resp.json(), [
            {'name': 'cursed-emojis'},
            {'name': 'dank-memes'},
        ])

    @override_settings(
        INGEST_S3_ACCESS_KEY='just-let-me-in',
        INGEST_S3_SECRET_KEY='i-dunno',
        INGEST_S3_ENDPOINT=None,
        INGEST_S3_REGION='atlantis-1',
        INGEST_EXTRA_BUCKETS=[
            'cute-plushies',
            'evil-mr-reviewers',
        ],
    )
    @mock_s3
    def test_list_buckets_extra(self):
        self.user.user_scopes.create(scope=Scope.S3Ingest)
        self.client.force_login(self.user)

        s3 = boto3.resource('s3', region_name='us-east-1')
        s3.create_bucket(Bucket='dank-memes')
        s3.create_bucket(Bucket='cursed-emojis')

        with self.assertNumQueries(3):
            resp = self.client.get(reverse('api:bucket-list'))
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.assertListEqual(resp.json(), [
            {'name': 'cursed-emojis'},
            {'name': 'cute-plushies'},
            {'name': 'dank-memes'},
            {'name': 'evil-mr-reviewers'},
        ])
