import uuid
from unittest.mock import MagicMock, call, patch

from django.test import override_settings
from django.urls import reverse
from django.utils import timezone
from rest_framework import status

from arkindex.documents.models import Corpus, ElementType
from arkindex.images.models import ImageServer
from arkindex.ponos.models import Farm, State, Task
from arkindex.process.models import (
    ActivityState,
    Process,
    ProcessMode,
    WorkerActivity,
    WorkerActivityState,
    WorkerVersion,
    WorkerVersionState,
)
from arkindex.process.utils import get_default_farm_id
from arkindex.project.tests import FixtureAPITestCase
from arkindex.training.models import Dataset, Model, ModelVersion, ModelVersionState
from arkindex.users.models import Role, User


class TestProcesses(FixtureAPITestCase):
    """
    Test process management
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.farm = Farm.objects.get(name='Wheat farm')
        cls.creds = cls.user.credentials.get()
        cls.repo = cls.creds.repos.get(url='http://my_repo.fake/workers/worker')
        cls.rev = cls.repo.revisions.get()
        cls.dataset1, cls.dataset2 = Dataset.objects.filter(corpus=cls.corpus).order_by('name')
        cls.private_corpus = Corpus.objects.create(name='Private corpus')
        cls.private_dataset = cls.private_corpus.datasets.create(
            name='Dead sea scrolls',
            description='Human instrumentality manual',
            creator=cls.user
        )
        cls.img_df = cls.corpus.files.create(
            name='test.jpg',
            size=42,
            content_type='image/jpg',
        )
        cls.pdf_df = cls.corpus.files.create(
            name='test.pdf',
            size=42,
            content_type='application/pdf',
        )
        cls.iiif_df = cls.corpus.files.create(
            name='test.json',
            size=42,
            content_type='application/json',
        )
        cls.page_type = ElementType.objects.get(corpus=cls.corpus, slug='page')
        cls.volume_type = ElementType.objects.get(corpus=cls.corpus, slug='volume')
        cls.ml_class = cls.corpus.ml_classes.create(name='clafoutis')
        cls.recognizer = WorkerVersion.objects.get(worker__slug='reco')
        cls.dla = WorkerVersion.objects.get(worker__slug='dla')
        cls.version_gpu = WorkerVersion.objects.get(worker__slug='worker-gpu')
        cls.workers_process = cls.corpus.processes.get(mode=ProcessMode.Workers)
        cls.version_with_model = WorkerVersion.objects.get(worker__slug='generic')

        # Create a model and a model version
        cls.model_1 = Model.objects.create(name='My model')
        cls.model_version_1 = ModelVersion.objects.create(model=cls.model_1, state=ModelVersionState.Available, size=8, hash='aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa', archive_hash='bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb')

        # Create multiple processes the user can access
        cls.user2 = User.objects.create_user('user2@test.test', display_name='Process creator')

        # Guest access (A user owns a process on a corpus they are not a member anymore)
        cls.private_corpus = Corpus.objects.create(name='Private')
        cls.user_img_process = Process.objects.create(
            mode=ProcessMode.Files,
            creator=cls.user,
            corpus=cls.private_corpus,
            farm=cls.farm,
        )
        cls.private_ml_class = cls.private_corpus.ml_classes.create(name='beignet')

        # Contributor access
        cls.repo.memberships.create(user=cls.user, level=Role.Contributor.value)
        cls.repository_process = Process.objects.create(
            mode=ProcessMode.Repository,
            creator=cls.user2,
            revision=cls.rev,
        )
        # Admin access
        cls.elts_process = cls.corpus.processes.create(mode=ProcessMode.Workers, creator=cls.user2)
        cls.processes = Process.objects.filter(
            id__in=[cls.user_img_process.id, cls.repository_process.id, cls.elts_process.id]
        )
        # Training process
        cls.train, cls.val, cls.test = (
            cls.corpus.elements.create(name=name, type=cls.volume_type)
            for name in ('train', 'val', 'test')
        )
        cls.training_model = Model.objects.create(name='Generic model', public=False)
        cls.training_process = Process.objects.create(
            creator=cls.user,
            corpus=cls.corpus,
            mode=ProcessMode.Training,
            model=cls.training_model,
            train_folder=cls.train,
            validation_folder=cls.val,
            test_folder=cls.test,
        )

        cls.local_process = Process.objects.get(
            creator=cls.user,
            mode=ProcessMode.Local,
        )
        cls.local_process_2 = Process.objects.create(
            creator=cls.user2,
            mode=ProcessMode.Local,
        )

        cls.import_worker_version = WorkerVersion.objects.get(worker__slug='file_import')

    def test_list_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.get(reverse('api:process-list'))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_list(self):
        """
        A user can list processes they have access to
        """
        # Order processes by causing updates on them
        self.elts_process.run()
        self.user_img_process.save()
        self.repository_process.save()
        # An update on a process task will make it first
        task = self.elts_process.tasks.first()
        task.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:process-list'))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        # Processes are ordered
        self.assertDictEqual(response.json(), {
            'count': 5,
            'number': 1,
            'next': None,
            'previous': None,
            'results': [
                {
                    'id': str(self.repository_process.id),
                    'name': None,
                    'mode': 'repository',
                    'corpus': None,
                    'state': 'unscheduled',
                    'activity_state': 'disabled',
                    'created': self.repository_process.created.isoformat().replace('+00:00', 'Z'),
                    'creator': 'Process creator',
                    'updated': self.repository_process.updated.isoformat().replace('+00:00', 'Z'),
                    'started': None,
                    'finished': None,
                },
                {
                    'id': str(self.user_img_process.id),
                    'name': None,
                    'mode': 'files',
                    'corpus': str(self.private_corpus.id),
                    'state': 'unscheduled',
                    'activity_state': 'disabled',
                    'created': self.user_img_process.created.isoformat().replace('+00:00', 'Z'),
                    'creator': 'Test user',
                    'updated': self.user_img_process.updated.isoformat().replace('+00:00', 'Z'),
                    'started': None,
                    'finished': None,
                },
                {
                    'id': str(self.elts_process.id),
                    'name': None,
                    'mode': 'workers',
                    'corpus': str(self.corpus.id),
                    'state': 'unscheduled',
                    'activity_state': 'disabled',
                    'created': self.elts_process.created.isoformat().replace('+00:00', 'Z'),
                    'creator': 'Process creator',
                    'updated': self.elts_process.updated.isoformat().replace('+00:00', 'Z'),
                    'started': self.elts_process.started.isoformat().replace('+00:00', 'Z'),
                    'finished': None,
                },
                {
                    'id': str(self.training_process.id),
                    'name': None,
                    'mode': 'training',
                    'corpus': str(self.corpus.id),
                    'state': 'unscheduled',
                    'activity_state': 'disabled',
                    'created': self.training_process.created.isoformat().replace('+00:00', 'Z'),
                    'creator': 'Test user',
                    'updated': self.training_process.updated.isoformat().replace('+00:00', 'Z'),
                    'started': None,
                    'finished': None,
                },
                {
                    'id': str(self.workers_process.id),
                    'name': 'Process fixture',
                    'mode': 'workers',
                    'corpus': str(self.corpus.id),
                    'state': 'unscheduled',
                    'activity_state': 'disabled',
                    'created': self.workers_process.created.isoformat().replace('+00:00', 'Z'),
                    'creator': 'Test user',
                    'updated': self.workers_process.updated.isoformat().replace('+00:00', 'Z'),
                    'started': None,
                    'finished': None,
                },
            ]
        })

    @override_settings(IMPORTS_WORKER_VERSION=None)
    def test_list_started(self):
        """
        Filter processes that have been started
        """
        with self.settings(IMPORTS_WORKER_VERSION=str(self.import_worker_version.id)):
            self.user_img_process.run()

        self.client.force_login(self.user)
        self.user_img_process.activity_state = ActivityState.Ready
        self.user_img_process.save()

        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:process-list'), {'started': 'true'})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            'count': 1,
            'number': 1,
            'next': None,
            'previous': None,
            'results': [
                {
                    'id': str(self.user_img_process.id),
                    'name': None,
                    'mode': 'files',
                    'corpus': str(self.private_corpus.id),
                    'state': 'unscheduled',
                    'activity_state': 'ready',
                    'created': self.user_img_process.created.isoformat().replace('+00:00', 'Z'),
                    'creator': 'Test user',
                    'updated': self.user_img_process.updated.isoformat().replace('+00:00', 'Z'),
                    'started': self.user_img_process.started.isoformat().replace('+00:00', 'Z'),
                    'finished': None,
                },
            ],
        })

    def test_list_exclude_started(self):
        """
        Filter process that have not been started yet
        """
        self.repository_process.run()
        self.elts_process.run()
        self.client.force_login(self.user)

        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:process-list'), {'started': 'false'})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            'count': 3,
            'number': 1,
            'next': None,
            'previous': None,
            'results': [
                {
                    'id': str(self.training_process.id),
                    'name': None,
                    'mode': 'training',
                    'corpus': str(self.corpus.id),
                    'state': 'unscheduled',
                    'activity_state': 'disabled',
                    'created': self.training_process.created.isoformat().replace('+00:00', 'Z'),
                    'creator': 'Test user',
                    'updated': self.training_process.updated.isoformat().replace('+00:00', 'Z'),
                    'started': None,
                    'finished': None,
                },
                {
                    'id': str(self.user_img_process.id),
                    'name': None,
                    'mode': 'files',
                    'corpus': str(self.private_corpus.id),
                    'state': 'unscheduled',
                    'activity_state': 'disabled',
                    'created': self.user_img_process.created.isoformat().replace('+00:00', 'Z'),
                    'creator': 'Test user',
                    'updated': self.user_img_process.updated.isoformat().replace('+00:00', 'Z'),
                    'started': None,
                    'finished': None,
                },
                {
                    'id': str(self.workers_process.id),
                    'name': 'Process fixture',
                    'mode': 'workers',
                    'corpus': str(self.corpus.id),
                    'state': 'unscheduled',
                    'activity_state': 'disabled',
                    'created': self.workers_process.created.isoformat().replace('+00:00', 'Z'),
                    'creator': 'Test user',
                    'updated': self.workers_process.updated.isoformat().replace('+00:00', 'Z'),
                    'started': None,
                    'finished': None,
                },
            ]
        })

    def test_list_filter_corpus(self):
        self.client.force_login(self.superuser)
        corpus2 = Corpus.objects.create(name='Another corpus', description='something')
        process2 = corpus2.processes.create(creator=self.user, mode=ProcessMode.Files)

        with self.assertNumQueries(4):
            response = self.client.get(reverse('api:process-list'), {'corpus': str(corpus2.id), 'started': False})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(process2.id))

    def test_list_filter_corpus_invalid(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(reverse('api:process-list'), {'corpus': 'oh-no'})
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {'corpus': ["'oh-no' is not a valid UUID"]})

    def test_list_filter_mode(self):
        self.client.force_login(self.user)
        assert Process.objects.exclude(mode=ProcessMode.Files).exists()

        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:process-list'), {'mode': ProcessMode.Files.value, 'started': False})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(self.user_img_process.id))

    def test_list_filter_mode_local(self):
        """
        Filtering by the local mode shows only the local process of the current user
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:process-list'), {'mode': ProcessMode.Local.value})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            'count': 1,
            'number': 1,
            'next': None,
            'previous': None,
            'results': [
                {
                    'id': str(self.local_process.id),
                    'name': None,
                    'mode': 'local',
                    'corpus': None,
                    'state': 'unscheduled',
                    'activity_state': 'disabled',
                    'created': self.local_process.created.isoformat().replace('+00:00', 'Z'),
                    'creator': 'Test user',
                    'updated': self.local_process.updated.isoformat().replace('+00:00', 'Z'),
                    'started': None,
                    'finished': None,
                },
            ]
        })

    def test_list_filter_unexisting_mode(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(reverse('api:process-list'), {'mode': 'unexisting_mode'})
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {'mode': ["Mode 'unexisting_mode' does not exist"]})

    def test_list_filter_created(self):
        """
        Display processes that have been created by the user
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:process-list'), {'created': 'true'})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            'count': 3,
            'number': 1,
            'next': None,
            'previous': None,
            'results': [
                {
                    'id': str(self.training_process.id),
                    'name': None,
                    'mode': 'training',
                    'corpus': str(self.corpus.id),
                    'state': 'unscheduled',
                    'activity_state': 'disabled',
                    'created': self.training_process.created.isoformat().replace('+00:00', 'Z'),
                    'creator': 'Test user',
                    'updated': self.training_process.updated.isoformat().replace('+00:00', 'Z'),
                    'started': None,
                    'finished': None,
                },
                {
                    'id': str(self.user_img_process.id),
                    'name': None,
                    'mode': 'files',
                    'corpus': str(self.private_corpus.id),
                    'state': 'unscheduled',
                    'activity_state': 'disabled',
                    'created': self.user_img_process.created.isoformat().replace('+00:00', 'Z'),
                    'creator': 'Test user',
                    'updated': self.user_img_process.updated.isoformat().replace('+00:00', 'Z'),
                    'started': None,
                    'finished': None,
                },
                {
                    'id': str(self.workers_process.id),
                    'name': 'Process fixture',
                    'mode': 'workers',
                    'corpus': str(self.corpus.id),
                    'state': 'unscheduled',
                    'activity_state': 'disabled',
                    'created': self.workers_process.created.isoformat().replace('+00:00', 'Z'),
                    'creator': 'Test user',
                    'updated': self.workers_process.updated.isoformat().replace('+00:00', 'Z'),
                    'started': None,
                    'finished': None,
                },
            ]
        })

    def test_list_filter_not_created(self):
        """
        Display processes that have not been created by the user
        """
        self.client.force_login(self.user)

        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:process-list'), {'created': 'false'})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(data['count'], 2)
        self.assertCountEqual(
            [p['id'] for p in data['results']],
            [str(p.id) for p in self.processes.exclude(creator=self.user)]
        )

    def test_list_filter_id(self):
        self.client.force_login(self.user)

        # Generate a UUID whose first characters are not the same as the first Process
        process_id = str(uuid.uuid4())
        while process_id[:10] == str(self.user_img_process.id)[:10]:
            process_id = str(uuid.uuid4())

        process2 = self.corpus.processes.create(
            id=process_id,
            creator=self.user,
            mode=ProcessMode.Files,
        )

        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:process-list'), {'id': process_id[:10], 'started': False})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(process2.id))

    def test_list_filter_name(self):
        self.client.force_login(self.user)

        process1 = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            name='Number One'
        )

        self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            name='Numero Duo'
        )

        with self.assertNumQueries(7):
            response = self.client.get(reverse('api:process-list'), {'name': 'Numb', 'started': False})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        data = response.json()
        self.assertEqual(len(data['results']), 1)
        self.assertEqual(data['results'][0]['id'], str(process1.id))

    def test_list_process_filter_wrong_state_value(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            response = self.client.get(reverse('api:process-list'), {'state': 'spain'})
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {'state': ["State 'spain' does not exist"]})

    def test_list_process_state_filter(self):
        """
        Only last run should be used to filter processes by state
        """
        completed_process = Process.objects.create(mode=ProcessMode.Workers, creator=self.user, corpus=self.corpus)
        completed_process.tasks.create(depth=0, run=0, slug='task', state=State.Failed)
        completed_process.tasks.create(depth=0, run=1, slug='task', state=State.Completed)
        self.assertEqual(completed_process.state, State.Completed)

        error_process = Process.objects.create(mode=ProcessMode.Workers, creator=self.user, corpus=self.corpus)
        error_process.tasks.create(depth=0, run=0, slug='error_task', state=State.Error)
        error_process.tasks.create(depth=0, run=0, slug='completed_task', state=State.Completed)
        self.assertEqual(error_process.state, State.Error)

        stopped_process = Process.objects.create(mode=ProcessMode.Workers, creator=self.user, corpus=self.corpus)
        stopped_process.tasks.create(depth=0, run=0, slug='completed_task', state=State.Completed)
        stopped_process.tasks.create(depth=0, run=1, slug='completed_task', state=State.Completed)
        stopped_process.tasks.create(depth=0, run=2, slug='stopped_task', state=State.Stopped)
        self.assertEqual(stopped_process.state, State.Stopped)

        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:process-list'), {'state': 'completed'})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertCountEqual([r['id'] for r in response.json()['results']], [str(completed_process.id)])

    def test_list_process_state_filter_stopped(self):
        stopping_process = Process.objects.create(mode=ProcessMode.Workers, creator=self.user, corpus=self.corpus)
        stopping_process.tasks.create(depth=0, run=0, slug='task', state=State.Error)
        stopping_process.tasks.create(depth=0, run=1, slug='task', state=State.Stopping)
        self.assertEqual(stopping_process.state, State.Stopping)

        error_process = Process.objects.create(mode=ProcessMode.Workers, creator=self.user, corpus=self.corpus)
        error_process.tasks.create(depth=0, run=0, slug='error_task', state=State.Error)
        error_process.tasks.create(depth=0, run=0, slug='stopping_task', state=State.Stopping)
        self.assertEqual(error_process.state, State.Error)

        stopped_process_1 = Process.objects.create(mode=ProcessMode.Workers, creator=self.user, corpus=self.corpus)
        stopped_process_1.tasks.create(depth=0, run=0, slug='unscheduled_task', state=State.Unscheduled)
        stopped_process_1.tasks.create(depth=0, run=0, slug='stopped_task', state=State.Stopped)
        self.assertEqual(stopped_process_1.state, State.Stopped)

        stopped_process_2 = Process.objects.create(mode=ProcessMode.Workers, creator=self.user, corpus=self.corpus)
        stopped_process_2.tasks.create(depth=0, run=0, slug='failed_task', state=State.Failed)
        stopped_process_2.tasks.create(depth=0, run=1, slug='pending_task', state=State.Pending)
        stopped_process_2.tasks.create(depth=0, run=1, slug='stopped_task', state=State.Stopped)
        self.assertEqual(stopped_process_2.state, State.Stopped)

        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:process-list'), {'state': 'stopped'})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertCountEqual(
            [r['id'] for r in response.json()['results']],
            [str(stopped_process_1.id), str(stopped_process_2.id)]
        )

    def test_list_process_state_filter_unscheduled(self):
        """
        Processes with no task should be interpreted as unscheduled
        """
        unscheduled_with_tasks = Process.objects.create(
            mode=ProcessMode.Workers,
            creator=self.user,
            corpus=self.corpus,
        )
        unscheduled_with_tasks.tasks.create(depth=0, run=0, slug='task', state=State.Completed)
        unscheduled_with_tasks.tasks.create(depth=0, run=1, slug='task', state=State.Unscheduled)
        unscheduled_with_tasks.started = timezone.now()
        unscheduled_with_tasks.save()
        self.assertEqual(unscheduled_with_tasks.state, State.Unscheduled)

        self.assertFalse(self.user_img_process.tasks.exists())
        self.assertEqual(self.user_img_process.state, State.Unscheduled)
        self.assertFalse(self.repository_process.tasks.exists())
        self.assertEqual(self.repository_process.state, State.Unscheduled)
        self.assertFalse(self.elts_process.tasks.exists())
        self.assertEqual(self.elts_process.state, State.Unscheduled)

        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:process-list'), {'state': 'unscheduled', 'started': True})
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertDictEqual(response.json(), {
            'count': 1,
            'number': 1,
            'next': None,
            'previous': None,
            'results': [
                {
                    'id': str(unscheduled_with_tasks.id),
                    'name': None,
                    'mode': 'workers',
                    'corpus': str(self.corpus.id),
                    'state': 'unscheduled',
                    'activity_state': 'disabled',
                    'created': unscheduled_with_tasks.created.isoformat().replace('+00:00', 'Z'),
                    'creator': 'Test user',
                    'updated': unscheduled_with_tasks.updated.isoformat().replace('+00:00', 'Z'),
                    'started': unscheduled_with_tasks.started.isoformat().replace('+00:00', 'Z'),
                    'finished': None,
                },
            ]
        })

        with self.assertNumQueries(4):
            response = self.client.get(reverse('api:process-list'), {'state': 'unscheduled', 'started': False})

        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertDictEqual(response.json(), {
            'count': 5,
            'number': 1,
            'next': None,
            'previous': None,
            'results': [
                {
                    'id': str(self.training_process.id),
                    'name': None,
                    'mode': 'training',
                    'corpus': str(self.corpus.id),
                    'state': 'unscheduled',
                    'activity_state': 'disabled',
                    'created': self.training_process.created.isoformat().replace('+00:00', 'Z'),
                    'creator': 'Test user',
                    'updated': self.training_process.updated.isoformat().replace('+00:00', 'Z'),
                    'started': None,
                    'finished': None,
                },
                {
                    'id': str(self.elts_process.id),
                    'name': None,
                    'mode': 'workers',
                    'corpus': str(self.corpus.id),
                    'state': 'unscheduled',
                    'activity_state': 'disabled',
                    'created': self.elts_process.created.isoformat().replace('+00:00', 'Z'),
                    'creator': 'Process creator',
                    'updated': self.elts_process.updated.isoformat().replace('+00:00', 'Z'),
                    'started': None,
                    'finished': None,
                },
                {
                    'id': str(self.repository_process.id),
                    'name': None,
                    'mode': 'repository',
                    'corpus': None,
                    'state': 'unscheduled',
                    'activity_state': 'disabled',
                    'created': self.repository_process.created.isoformat().replace('+00:00', 'Z'),
                    'creator': 'Process creator',
                    'updated': self.repository_process.updated.isoformat().replace('+00:00', 'Z'),
                    'started': None,
                    'finished': None,
                },
                {
                    'id': str(self.user_img_process.id),
                    'name': None,
                    'mode': 'files',
                    'corpus': str(self.private_corpus.id),
                    'state': 'unscheduled',
                    'activity_state': 'disabled',
                    'created': self.user_img_process.created.isoformat().replace('+00:00', 'Z'),
                    'creator': 'Test user',
                    'updated': self.user_img_process.updated.isoformat().replace('+00:00', 'Z'),
                    'started': None,
                    'finished': None,
                },
                {
                    'id': str(self.workers_process.id),
                    'name': 'Process fixture',
                    'mode': 'workers',
                    'corpus': str(self.corpus.id),
                    'state': 'unscheduled',
                    'activity_state': 'disabled',
                    'created': self.workers_process.created.isoformat().replace('+00:00', 'Z'),
                    'creator': 'Test user',
                    'updated': self.workers_process.updated.isoformat().replace('+00:00', 'Z'),
                    'started': None,
                    'finished': None,
                },
            ]
        })

    def test_details_requires_login(self):
        response = self.client.get(reverse('api:process-details', kwargs={'pk': self.user_img_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    @override_settings(IMPORTS_WORKER_VERSION=None)
    def test_details(self):
        self.client.force_login(self.user)
        # A user needs access to a process's corpus to get this process's details
        self.private_corpus.memberships.create(user=self.user, level=Role.Guest.value)
        # Have the endpoint return a task too
        with self.settings(IMPORTS_WORKER_VERSION=str(self.import_worker_version.id)):
            self.user_img_process.run()
        task = self.user_img_process.tasks.get()

        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:process-details', kwargs={'pk': self.user_img_process.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.json(), {
            'activity_state': ActivityState.Disabled.value,
            'corpus': str(self.private_corpus.id),
            'element': None,
            'element_name_contains': None,
            'element_type': None,
            'farm': {
                'id': str(self.farm.id),
                'name': 'Wheat farm',
            },
            'files': [],
            'folder_type': None,
            'id': str(self.user_img_process.id),
            'load_children': False,
            'ml_class_id': None,
            'mode': 'files',
            'model_id': None,
            'name': None,
            'revision': None,
            'state': 'unscheduled',
            'tasks': [
                {
                    'id': str(task.id),
                    'depth': 0,
                    'parents': [],
                    'run': 0,
                    'shm_size': None,
                    'slug': 'import_files',
                    'state': 'unscheduled',
                }
            ],
            'template_id': None,
            'test_folder_id': None,
            'train_folder_id': None,
            'use_cache': False,
            'use_gpu': False,
            'validation_folder_id': None,
            'created': self.user_img_process.created.isoformat().replace('+00:00', 'Z'),
            'updated': self.user_img_process.updated.isoformat().replace('+00:00', 'Z'),
            'started': self.user_img_process.started.isoformat().replace('+00:00', 'Z'),
            'finished': None,
        })

    def test_details_training_process(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.get(reverse('api:process-details', kwargs={'pk': self.training_process.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.assertEqual(response.json(), {
            'activity_state': ActivityState.Disabled.value,
            'corpus': str(self.corpus.id),
            'element': None,
            'element_name_contains': None,
            'element_type': None,
            'farm': None,
            'files': [],
            'folder_type': None,
            'id': str(self.training_process.id),
            'load_children': False,
            'ml_class_id': None,
            'mode': ProcessMode.Training.value,
            'model_id': str(self.training_model.id),
            'name': None,
            'revision': None,
            'state': 'unscheduled',
            'tasks': [],
            'template_id': None,
            'test_folder_id': str(self.test.id),
            'train_folder_id': str(self.train.id),
            'use_cache': False,
            'use_gpu': False,
            'validation_folder_id': str(self.val.id),
            'created': self.training_process.created.isoformat().replace('+00:00', 'Z'),
            'updated': self.training_process.updated.isoformat().replace('+00:00', 'Z'),
            'started': None,
            'finished': None,
        })

    def test_delete_requires_login(self):
        response = self.client.delete(reverse('api:process-details', kwargs={'pk': self.user_img_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_delete_requires_verified(self):
        user = User.objects.create(email='AAAAAAA')
        self.client.force_login(user)
        response = self.client.delete(reverse('api:process-details', kwargs={'pk': self.user_img_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have permission to perform this action.'})

    def test_delete_running(self):
        """
        It is not possible to delete a running import
        """
        self.client.force_login(self.user)
        self.elts_process.run()
        self.elts_process.tasks.update(state=State.Running)
        response = self.client.delete(reverse('api:process-details', kwargs={'pk': self.elts_process.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'__all__': ['Cannot delete a process while it is running']})

    def test_delete_activities_pending(self):
        """
        It is not possible to delete an import with non initialized activities
        """
        self.client.force_login(self.user)
        self.elts_process.run()
        self.elts_process.activity_state = ActivityState.Pending
        self.elts_process.save()
        response = self.client.delete(reverse('api:process-details', kwargs={'pk': self.elts_process.id}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'__all__': ['Please wait activities to be initialized before deleting this process']}
        )

    def test_delete_repository_import_no_permission(self):
        """
        Deletion of a repository import requires to be admin on the repository
        """
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:process-details', kwargs={'pk': self.repository_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have a sufficient access level to this process.'})

    def test_delete_corpus_import_no_permission(self):
        """
        A user cannot delete a process linked to a corpus he has no admin access to
        """
        self.client.force_login(self.user)
        self.assertFalse(self.user_img_process.corpus.memberships.filter(user=self.user).exists())
        response = self.client.delete(reverse('api:process-details', kwargs={'pk': self.user_img_process.id}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertDictEqual(response.json(), {'detail': 'Not found.'})

    def test_delete_corpus_import_admin(self):
        """
        A user is allowed to delete a process if he has an admin right to its corpus.
        If no activities exists for this process, it is deleted directly.
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(16):
            response = self.client.delete(reverse('api:process-details', kwargs={'pk': self.elts_process.id}))
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(Process.DoesNotExist):
            self.elts_process.refresh_from_db()

    def test_delete_superuser(self):
        """
        A superuser is allowed to delete any process
        """
        self.client.force_login(self.superuser)
        with self.assertNumQueries(13):
            response = self.client.delete(reverse('api:process-details', kwargs={'pk': self.user_img_process.id}))
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        with self.assertRaises(Process.DoesNotExist):
            self.user_img_process.refresh_from_db()

    @patch('arkindex.project.triggers.process_tasks.process_delete.delay')
    def test_delete_corpus_import_async(self, delay_mock):
        """
        A process with activities is deleted in an async task and a HTTP_202_ACCEPTED is returned
        """
        self.client.force_login(self.user)
        self.elts_process.activities.create(
            element=self.corpus.elements.get(name='Volume 1'),
            worker_version=WorkerVersion.objects.get(worker__slug='reco'),
        )
        with self.assertNumQueries(8):
            response = self.client.delete(reverse('api:process-details', kwargs={'pk': self.elts_process.id}))
            self.assertEqual(response.status_code, status.HTTP_202_ACCEPTED)
        self.assertEqual(delay_mock.call_count, 1)
        (process_arg,), kwargs = delay_mock.call_args
        self.assertEqual(process_arg.id, self.elts_process.id)
        self.assertEqual(kwargs, {'delete_activity': WorkerActivityState.Queued})

    @patch('arkindex.project.triggers.process_tasks.process_delete.delay')
    def test_delete_process_activities(self, delay_mock):
        """
        Deleting a process deletes the related activities with a queued state and
        sets the the FK on remaining activities to null
        Mock the async deletion task to run it directly in this test
        """
        from arkindex.process.tasks import process_delete
        delay_mock.side_effect = process_delete
        self.client.force_login(self.superuser)
        queued_activity = WorkerActivity.objects.create(
            element=self.corpus.elements.get(name='Volume 1'),
            process=self.elts_process,
            worker_version=WorkerVersion.objects.get(worker__slug='reco'),
            state=WorkerActivityState.Queued,
        )
        completed_activity = WorkerActivity.objects.create(
            element=self.corpus.elements.get(name='Volume 1, page 2r'),
            process=self.elts_process,
            worker_version=WorkerVersion.objects.get(worker__slug='reco'),
            state=WorkerActivityState.Processed,
        )
        with self.assertNumQueries(14):
            response = self.client.delete(reverse('api:process-details', kwargs={'pk': self.elts_process.id}))
            self.assertEqual(response.status_code, status.HTTP_202_ACCEPTED)
        with self.assertRaises(Process.DoesNotExist):
            self.elts_process.refresh_from_db()
        with self.assertRaises(WorkerActivity.DoesNotExist):
            queued_activity.refresh_from_db()
        with self.assertRaises(WorkerActivity.DoesNotExist):
            completed_activity.refresh_from_db()

    def test_partial_update_requires_login(self):
        response = self.client.patch(reverse('api:process-details', kwargs={'pk': self.user_img_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_partial_update_requires_verified(self):
        user = User.objects.create(email="unchecked")
        self.client.force_login(user)
        response = self.client.patch(reverse('api:process-details', kwargs={'pk': self.user_img_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have permission to perform this action.'})

    def test_partial_update_running(self):
        """
        A running process cannot be updated... (see following test)
        """
        self.client.force_login(self.user)
        self.elts_process.run()
        self.elts_process.tasks.update(state=State.Running)
        with self.assertNumQueries(9):
            response = self.client.patch(
                reverse('api:process-details', kwargs={'pk': self.elts_process.id}),
                {'element_name_contains': 'something'},
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'__all__': ['Cannot edit a process while it is running']})

    def test_partial_update_running_only_name(self):
        """
        Only name attribute of a running process can be patched
        """
        self.client.force_login(self.user)
        self.elts_process.run()
        self.elts_process.tasks.update(state=State.Running)
        self.assertEqual(self.elts_process.name, None)
        with self.assertNumQueries(14):
            response = self.client.patch(
                reverse('api:process-details', kwargs={'pk': self.elts_process.id}),
                {'name': 'newName'},
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.elts_process.refresh_from_db()
        self.assertEqual(self.elts_process.name, 'newName')

    @override_settings(IMPORTS_WORKER_VERSION=None)
    def test_partial_update_file_import_element(self):
        """
        A file import's element can be updated even while it is running
        """
        self.client.force_login(self.user)
        process = self.corpus.processes.create(mode=ProcessMode.Files, creator=self.user)
        with self.settings(IMPORTS_WORKER_VERSION=str(self.import_worker_version.id)):
            process.run()
            process.tasks.update(state=State.Running)
        self.assertIsNone(self.elts_process.element)

        # Take any element with an image: this will cause extra queries to retrieve the Image and ImageServer
        element = self.corpus.elements.exclude(image_id=None).first()

        with self.assertNumQueries(15), self.settings(IMPORTS_WORKER_VERSION=str(self.import_worker_version.id)):
            response = self.client.patch(
                reverse('api:process-details', kwargs={'pk': process.id}),
                {'element_id': str(element.id)},
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        process.refresh_from_db()
        self.assertEqual(process.element, element)

    def test_partial_update_no_permission(self):
        """
        A user cannot update a process linked to a corpus he has no admin access to
        """
        self.client.force_login(self.user)
        self.corpus.memberships.filter(user=self.user).update(level=Role.Guest.value)
        response = self.client.patch(reverse('api:process-details', kwargs={'pk': self.elts_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have a sufficient access level to this process.'})

    def test_partial_update_repository_requires_admin(self):
        """
        Edition of a repository import requires to be admin on the repository
        """
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:process-details', kwargs={'pk': self.repository_process.id}),
            {'name': 'newName'}
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have a sufficient access level to this process.'})

    def test_partial_update_stop(self):
        """
        A pending or running process can be stopped
        """
        self.client.force_login(self.user)
        self.elts_process.run()

        for state in [State.Pending, State.Running]:
            with self.subTest(state=state):
                self.elts_process.tasks.update(state=state)
                self.elts_process.refresh_from_db()
                self.assertEqual(self.elts_process.state, state)

                with self.assertNumQueries(16):
                    response = self.client.patch(
                        reverse('api:process-details', kwargs={'pk': self.elts_process.id}),
                        {'state': 'stopping'},
                        format='json',
                    )
                    self.assertEqual(response.status_code, status.HTTP_200_OK)

                self.elts_process.refresh_from_db()
                self.assertEqual(self.elts_process.state, State.Stopping)
                self.assertIsNone(self.elts_process.finished)

    def test_partial_update_unscheduled_stop(self):
        """
        An unscheduled process can be stopped
        """
        self.client.force_login(self.user)
        self.elts_process.run()
        self.assertEqual(self.elts_process.state, State.Unscheduled)

        with self.assertNumQueries(17):
            response = self.client.patch(
                reverse('api:process-details', kwargs={'pk': self.elts_process.id}),
                {'state': 'stopping'},
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.elts_process.refresh_from_db()
        # Since no agents need to do anything to stop the tasks, all the tasks are immediately marked as Stopped
        self.assertEqual(self.elts_process.state, State.Stopped)
        # Since everything is stopped, `finished` should be set
        self.assertIsNotNone(self.elts_process.finished)

    def test_partial_update_state_only_stopping(self):
        """
        A process' state can only be updated to `stopping`
        """
        self.client.force_login(self.user)
        self.elts_process.run()
        self.assertEqual(self.elts_process.state, State.Unscheduled)

        for state in set(State) - {State.Stopping}:
            with self.subTest(state=state):
                with self.assertNumQueries(9):
                    response = self.client.patch(
                        reverse('api:process-details', kwargs={'pk': self.elts_process.id}),
                        {'state': state.value},
                        format='json',
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertDictEqual(response.json(), {
                    'state': ["Can only change the state to 'stopping'"],
                })

                self.elts_process.refresh_from_db()
                self.assertEqual(self.elts_process.state, State.Unscheduled)

    def test_partial_update_state_non_final(self):
        """
        Only a process in an unscheduled, pending or running state can be updated to `stopping`
        """
        self.client.force_login(self.user)
        self.elts_process.run()

        for state in set(State) - {State.Unscheduled, State.Pending, State.Running}:
            with self.subTest(state=state):
                self.elts_process.tasks.update(state=state)
                self.elts_process.refresh_from_db()
                self.assertEqual(self.elts_process.state, state)

                with self.assertNumQueries(9):
                    response = self.client.patch(
                        reverse('api:process-details', kwargs={'pk': self.elts_process.id}),
                        {'state': 'stopping'},
                        format='json',
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertDictEqual(response.json(), {
                    'state': [f'Cannot stop a {state.value} process'],
                })

                self.elts_process.refresh_from_db()
                self.assertEqual(self.elts_process.state, state)

    def test_update_wrong_mode(self):
        """
        Processes that are not of Workers mode cannot be updated, except the 'name' field
        """
        self.client.force_login(self.superuser)
        forbidden_modes = list(ProcessMode)
        forbidden_modes.remove(ProcessMode.Workers)
        # Local processes cannot be accessed by anyone because they have no corpus and no revision,
        # not even instance admins, so they cause a 404 instead of 400
        forbidden_modes.remove(ProcessMode.Local)
        # Tested separately; this process mode does not have a corpus, which causes other errors.
        forbidden_modes.remove(ProcessMode.Repository)

        for mode in forbidden_modes:
            with self.subTest(mode=mode):
                process = Process.objects.create(
                    mode=mode,
                    creator=self.user,
                    corpus=self.corpus,
                )

                response = self.client.put(
                    reverse('api:process-details', kwargs={'pk': process.id}),
                    {
                        'name': 'newName',
                        'template_id': None,
                        'element_name_contains': 'AAA',
                        'element_type': 'page',
                        'load_children': True
                    },
                    format='json'
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(
                    response.json(),
                    {'__all__': ['Only processes of mode Workers can be updated']}
                )

    def test_update_repository(self):
        """
        Trying to update any field other than the `name` field on a Repository process is not allowed
        """
        self.client.force_login(self.superuser)
        process = Process.objects.create(
            creator=self.user,
            mode=ProcessMode.Repository,
            revision=self.rev,
        )
        with self.assertNumQueries(5):
            response = self.client.put(
                reverse('api:process-details', kwargs={'pk': process.id}),
                {
                    'name': 'newName',
                    'template_id': None,
                    'element_type': 'page',
                    'element_name_contains': 'AAA',
                    'load_children': True,
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'element_type': ['Object with slug=page does not exist.']},
        )

    def test_update_unscheduled_stop(self):
        """
        An unscheduled process can be stopped
        """
        self.client.force_login(self.user)
        self.elts_process.run()
        self.assertEqual(self.elts_process.state, State.Unscheduled)

        with self.assertNumQueries(18):
            response = self.client.put(
                reverse('api:process-details', kwargs={'pk': self.elts_process.id}),
                {
                    'name': 'newName',
                    'template_id': None,
                    'element_name_contains': 'AAA',
                    'element_type': 'page',
                    'load_children': True,
                    'state': 'stopping',
                },
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.elts_process.refresh_from_db()
        # Since no agents need to do anything to stop the tasks, all the tasks are immediately marked as Stopped
        self.assertEqual(self.elts_process.state, State.Stopped)
        # Since everything is stopped, `finished` should be set
        self.assertIsNotNone(self.elts_process.finished)

    def test_update_state_only_stopping(self):
        """
        A process' state can only be updated to `stopping`
        """
        self.client.force_login(self.user)
        self.elts_process.run()
        self.assertEqual(self.elts_process.state, State.Unscheduled)

        for state in set(State) - {State.Stopping}:
            with self.subTest(state=state):
                with self.assertNumQueries(10):
                    response = self.client.put(
                        reverse('api:process-details', kwargs={'pk': self.elts_process.id}),
                        {
                            'name': 'newName',
                            'template_id': None,
                            'element_name_contains': 'AAA',
                            'element_type': 'page',
                            'load_children': True,
                            'state': state.value,
                        },
                        format='json',
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertDictEqual(response.json(), {
                    'state': ["Can only change the state to 'stopping'"],
                })

                self.elts_process.refresh_from_db()
                self.assertEqual(self.elts_process.state, State.Unscheduled)

    def test_update_state_non_final(self):
        """
        Only a process in an unscheduled, pending or running state can be updated to `stopping`
        """
        self.client.force_login(self.user)
        self.elts_process.run()

        for state in set(State) - {State.Unscheduled, State.Pending, State.Running}:
            with self.subTest(state=state):
                self.elts_process.tasks.update(state=state)
                self.elts_process.refresh_from_db()
                self.assertEqual(self.elts_process.state, state)

                with self.assertNumQueries(10):
                    response = self.client.put(
                        reverse('api:process-details', kwargs={'pk': self.elts_process.id}),
                        {
                            'name': 'newName',
                            'template_id': None,
                            'element_name_contains': 'AAA',
                            'element_type': 'page',
                            'load_children': True,
                            'state': 'stopping',
                        },
                        format='json',
                    )
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertDictEqual(response.json(), {
                    'state': [f'Cannot stop a {state.value} process'],
                })

                self.elts_process.refresh_from_db()
                self.assertEqual(self.elts_process.state, state)

    def test_update(self):
        """
        Process fields related to elements filtering can be updated
        """
        self.client.force_login(self.user)
        process = Process.objects.create(mode=ProcessMode.Workers, corpus=self.corpus, creator=self.user)
        self.assertIsNone(process.name)
        self.assertIsNone(process.name_contains)
        self.assertIsNone(process.element_type)
        self.assertIsNone(process.ml_class)
        self.assertFalse(process.load_children)

        with self.assertNumQueries(13):
            response = self.client.put(
                reverse('api:process-details', kwargs={'pk': process.id}),
                {
                    'name': 'newName',
                    'element_name_contains': 'AAA',
                    'element_type': 'page',
                    'ml_class_id': str(self.ml_class.id),
                    'load_children': True,
                    'template_id': None,
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        process.refresh_from_db()
        self.assertEqual(process.name, 'newName')
        self.assertEqual(process.name_contains, 'AAA')
        self.assertEqual(process.element_type, self.page_type)
        self.assertEqual(process.ml_class, self.ml_class)
        self.assertTrue(process.load_children)

    def test_update_invalid_ml_class(self):
        self.client.force_login(self.user)
        process = Process.objects.create(mode=ProcessMode.Workers, corpus=self.corpus, creator=self.user)

        non_existent_id = uuid.uuid4()
        cases = [
            ('je says pas', '“je says pas” is not a valid UUID.', 9),
            (non_existent_id, f'Invalid pk "{non_existent_id}" - object does not exist.', 10),
            (self.private_ml_class.id, f'Invalid pk "{self.private_ml_class.id}" - object does not exist.', 10),
        ]
        for ml_class_id, message, queries in cases:
            with self.subTest(ml_class_id=ml_class_id), self.assertNumQueries(queries):
                response = self.client.put(
                    reverse('api:process-details', kwargs={'pk': process.id}),
                    {
                        'name': 'newName',
                        'element_name_contains': 'AAA',
                        'element_type': 'page',
                        'ml_class_id': str(ml_class_id),
                        'load_children': True,
                        'template_id': None,
                    },
                    format='json'
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), {
                    'ml_class_id': [message],
                })

    def test_update_fields(self):
        """
        Ensure determined fields can be updated on a non started workers process.
        Editable fields are `name`, `name_contains`, `element_type`, and `load_children`.
        """
        self.client.force_login(self.user)
        process = Process.objects.create(mode=ProcessMode.Workers, corpus=self.corpus, creator=self.user)

        with self.assertNumQueries(12):
            response = self.client.put(
                reverse('api:process-details', kwargs={'pk': process.id}),
                {
                    'id': str(uuid.uuid4()),
                    'revision': str(uuid.uuid4()),
                    'element': str(uuid.uuid4()),
                    'workflow': str(uuid.uuid4()),
                    'creator': str(uuid.uuid4()),
                    'files': [1, 2, 3],
                    'mode': 'iiif',
                    'folder_type': 'register',
                    'name': 'newName',
                    'element_name_contains': 'AAA',
                    'element_type': 'page',
                    'load_children': True,
                    'use_cache': True,
                    'use_gpu': False,
                    'activity_state': ActivityState.Pending.value,
                    'template_id': None,
                    'created': '2038-01-19T03:14:08Z',
                    'updated': '2038-01-19T03:14:08Z',
                    'started': '2038-01-19T03:14:08Z',
                    'finished': '2038-01-19T03:14:08Z',
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK, response.json())

        process.refresh_from_db()
        self.assertEqual(response.json(), {
            'name': 'newName',
            'element_name_contains': 'AAA',
            'element_type': 'page',
            'folder_type': None,
            'ml_class_id': None,
            'load_children': True,
            'use_cache': False,
            'use_gpu': False,
            'activity_state': ActivityState.Disabled.value,
            'id': str(process.id),
            'corpus': str(self.corpus.id),
            'element': None,
            'files': [],
            'mode': 'workers',
            'revision': None,
            'state': 'unscheduled',
            'template_id': None,
            'model_id': None,
            'train_folder_id': None,
            'validation_folder_id': None,
            'test_folder_id': None,
            'farm': None,
            'tasks': [],
            'created': process.created.isoformat().replace('+00:00', 'Z'),
            'updated': process.updated.isoformat().replace('+00:00', 'Z'),
            'started': None,
            'finished': None,
        })

    def test_partial_update_wrong_mode(self):
        """
        Fields other than 'name' cannot be updated in processes that are not of Workers mode
        """
        self.client.force_login(self.superuser)
        forbidden_modes = list(ProcessMode)
        forbidden_modes.remove(ProcessMode.Workers)
        # Local processes cannot be accessed by anyone because they have no corpus and no revision,
        # not even instance admins, so they cause a 404 instead of 400
        forbidden_modes.remove(ProcessMode.Local)

        for mode in forbidden_modes:
            with self.subTest(mode=mode):
                process = Process(
                    mode=mode,
                    creator=self.user,
                    corpus=self.corpus,
                )
                if mode == ProcessMode.Repository:
                    process.corpus = None
                    process.revision = self.rev
                process.save()

                response = self.client.patch(
                    reverse('api:process-details', kwargs={'pk': process.id}),
                    {'element_name_contains': 'something'},
                    format='json',
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(
                    response.json(),
                    {'__all__': ['Only processes of mode Workers can be updated']}
                )

    def test_partial_update_only_name(self):
        """
        Whatever the process mode, except Local, name attribute (alone) can be updated
        """
        self.client.force_login(self.superuser)
        modes = list(ProcessMode)
        # Local processes cannot be accessed by anyone because they have no corpus and no revision,
        # not even instance admins, so they cause a 404
        modes.remove(ProcessMode.Local)

        for mode in modes:
            with self.subTest(mode=mode):
                process = Process(
                    mode=mode,
                    creator=self.user,
                    corpus=self.corpus,
                )
                if mode == ProcessMode.Repository:
                    process.corpus = None
                    process.revision = self.rev
                process.save()

                self.assertEqual(process.name, None)
                response = self.client.patch(
                    reverse('api:process-details', kwargs={'pk': process.id}),
                    {
                        'name': 'newName'
                    },
                    format='json'
                )
                self.assertEqual(response.status_code, status.HTTP_200_OK)
                process.refresh_from_db()
                self.assertEqual(process.name, 'newName')

    def test_partial_update_corpus_no_write_right(self):
        self.client.force_login(self.user)
        self.corpus.memberships.filter(user=self.user).update(level=Role.Guest.value)
        response = self.client.patch(reverse('api:process-details', kwargs={'pk': self.elts_process.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have a sufficient access level to this process.'})

    def test_partial_update(self):
        """
        Process fields related to elements filtering can be updated
        """
        self.client.force_login(self.user)
        process = Process.objects.create(mode=ProcessMode.Workers, corpus=self.corpus, creator=self.user)
        self.assertIsNone(process.name)
        self.assertIsNone(process.name_contains)
        self.assertIsNone(process.element_type)
        self.assertIsNone(process.ml_class)
        self.assertFalse(process.load_children)

        with self.assertNumQueries(13):
            response = self.client.patch(
                reverse('api:process-details', kwargs={'pk': process.id}),
                {
                    'name': 'newName',
                    'element_name_contains': 'AAA',
                    'element_type': 'page',
                    'ml_class_id': str(self.ml_class.id),
                    'load_children': True,
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        process.refresh_from_db()
        self.assertEqual(process.name, 'newName')
        self.assertEqual(process.name_contains, 'AAA')
        self.assertEqual(process.element_type, self.page_type)
        self.assertEqual(process.ml_class, self.ml_class)
        self.assertTrue(process.load_children)

    def test_partial_update_reset_fields(self):
        self.client.force_login(self.user)
        process = Process.objects.create(
            mode=ProcessMode.Workers,
            corpus=self.corpus,
            creator=self.user,
            name='newName',
            name_contains='AAA',
            element_type=self.page_type,
            load_children=True
        )
        response = self.client.patch(
            reverse('api:process-details', kwargs={'pk': process.id}),
            {
                'name': None,
                'element_name_contains': None,
                'element_type': None,
                'load_children': False
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        process.refresh_from_db()
        self.assertTupleEqual(
            (process.name, process.name_contains, process.element_type, process.load_children),
            (None, None, None, False)
        )

    def test_partial_update_wrong_type(self):
        self.client.force_login(self.user)
        process = Process.objects.create(mode=ProcessMode.Workers, corpus=self.corpus, creator=self.user)
        response = self.client.patch(
            reverse('api:process-details', kwargs={'pk': process.id}),
            {'element_type': 'unexisting_type'},
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {'element_type': ['Object with slug=unexisting_type does not exist.']}
        )

    def test_partial_update_invalid_ml_class(self):
        self.client.force_login(self.user)
        process = Process.objects.create(mode=ProcessMode.Workers, corpus=self.corpus, creator=self.user)

        non_existent_id = uuid.uuid4()
        cases = [
            ('je says pas', '“je says pas” is not a valid UUID.', 8),
            (non_existent_id, f'Invalid pk "{non_existent_id}" - object does not exist.', 9),
            (self.private_ml_class.id, f'Invalid pk "{self.private_ml_class.id}" - object does not exist.', 9),
        ]
        for ml_class_id, message, queries in cases:
            with self.subTest(ml_class_id=ml_class_id), self.assertNumQueries(queries):
                response = self.client.patch(
                    reverse('api:process-details', kwargs={'pk': process.id}),
                    {
                        'ml_class_id': str(ml_class_id),
                    },
                    format='json'
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), {
                    'ml_class_id': [message],
                })

    def test_partial_update_fields(self):
        """
        Ensure determined fields can be updated on a non started workers process.
        Editable fields are `name`, `name_contains`, `element_type`, and `load_children`.
        """
        self.client.force_login(self.user)
        process = Process.objects.create(mode=ProcessMode.Workers, corpus=self.corpus, creator=self.user)

        with self.assertNumQueries(12):
            response = self.client.patch(
                reverse('api:process-details', kwargs={'pk': process.id}),
                {
                    'id': str(uuid.uuid4()),
                    'revision': str(uuid.uuid4()),
                    'element': str(uuid.uuid4()),
                    'workflow': str(uuid.uuid4()),
                    'creator': str(uuid.uuid4()),
                    'files': [1, 2, 3],
                    'mode': 'iiif',
                    'folder_type': 'register',
                    'name': 'newName',
                    'element_name_contains': 'AAA',
                    'element_type': 'page',
                    'load_children': True,
                    'use_cache': True,
                    'use_gpu': False,
                    'activity_state': ActivityState.Pending.value,
                    'created': '2038-01-19T03:14:08Z',
                    'updated': '2038-01-19T03:14:08Z',
                    'started': '2038-01-19T03:14:08Z',
                    'finished': '2038-01-19T03:14:08Z',
                },
                format='json'
            )
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        process.refresh_from_db()
        self.assertEqual(response.json(), {
            'name': 'newName',
            'element_name_contains': 'AAA',
            'element_type': 'page',
            'ml_class_id': None,
            'load_children': True,
            'use_cache': False,
            'use_gpu': False,
            'activity_state': ActivityState.Disabled.value,
            'id': str(process.id),
            'corpus': str(self.corpus.id),
            'element': None,
            'files': [],
            'mode': 'workers',
            'revision': None,
            'state': 'unscheduled',
            'template_id': None,
            'model_id': None,
            'train_folder_id': None,
            'validation_folder_id': None,
            'test_folder_id': None,
            'farm': None,
            'tasks': [],
            'created': process.created.isoformat().replace('+00:00', 'Z'),
            'updated': process.updated.isoformat().replace('+00:00', 'Z'),
            'started': None,
            'finished': None,
        })

    def test_retry_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(reverse('api:process-retry', kwargs={'pk': self.user_img_process.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_retry_repository_process_no_right(self):
        """
        A user that is not the creator nor admin cannot restart a process that is not linked to any corpus
        """
        self.client.force_login(self.user)
        self.repository_process.run()
        self.repository_process.tasks.all().update(state=State.Error)
        self.assertEqual(self.repository_process.state, State.Error)

        with self.assertNumQueries(9):
            response = self.client.post(reverse('api:process-retry', kwargs={'pk': self.repository_process.id}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

        self.assertDictEqual(response.json(), {'detail': 'You do not have an admin access to this process.'})

    def test_retry_only_final(self):
        self.client.force_login(self.user)
        self.elts_process.run()
        parameters = [
            (State.Unscheduled, 'This process is already pending'),
            (State.Pending, 'This process is already pending'),
            (State.Running, 'This process is already running'),
            (State.Stopping, 'This process is stopping'),
        ]
        for state, message in parameters:
            with self.subTest(state=state):
                self.elts_process.tasks.all().update(state=state)
                self.assertEqual(self.elts_process.state, state)
                with self.assertNumQueries(9):
                    response = self.client.post(reverse('api:process-retry', kwargs={'pk': self.elts_process.id}))
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), {'__all__': [message]})

    @patch('arkindex.project.triggers.process_tasks.initialize_activity.delay')
    def test_retry(self, delay_mock):
        self.client.force_login(self.user)
        self.elts_process.run()
        self.elts_process.tasks.all().update(state=State.Error)
        self.assertEqual(self.elts_process.state, State.Error)
        self.elts_process.finished = timezone.now()
        self.elts_process.save()

        with self.assertNumQueries(14):
            response = self.client.post(reverse('api:process-retry', kwargs={'pk': self.elts_process.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.elts_process.refresh_from_db()
        self.assertEqual(self.elts_process.state, State.Unscheduled)
        # `finished` is unset
        self.assertIsNone(self.elts_process.finished)
        # Activity initialization runs again
        self.assertFalse(delay_mock.called)

    @patch('arkindex.project.triggers.process_tasks.initialize_activity.delay')
    def test_retry_no_tasks(self, delay_mock):
        self.client.force_login(self.user)
        self.assertFalse(self.elts_process.tasks.exists())

        with self.assertNumQueries(14):
            response = self.client.post(reverse('api:process-retry', kwargs={'pk': self.elts_process.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.elts_process.refresh_from_db()
        self.assertEqual(self.elts_process.state, State.Unscheduled)
        self.assertTrue(self.elts_process.tasks.exists())
        self.assertFalse(delay_mock.called)

    def test_retry_repo_disabled(self):
        self.repository_process.revision = self.rev
        self.repository_process.save()
        self.repository_process.run()
        self.repository_process.tasks.all().update(state=State.Error)
        self.assertEqual(self.repository_process.state, State.Error)
        self.creds.delete()
        # Allow the user to do the retry
        self.repo.memberships.filter(user=self.user).update(level=Role.Admin.value)
        self.client.force_login(self.user)

        with self.assertNumQueries(8):
            response = self.client.post(reverse('api:process-retry', kwargs={'pk': self.repository_process.id}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {'__all__': ['Git repository does not have any valid credentials']})

    @patch('arkindex.project.triggers.process_tasks.initialize_activity.delay')
    def test_retry_initialize_activity(self, delay_mock):
        """
        Retrying a process should re-initialize the WorkerActivities if they are enabled
        """
        self.client.force_login(self.user)
        self.workers_process.run()
        self.workers_process.tasks.all().update(state=State.Error)
        self.assertEqual(self.workers_process.state, State.Error)

        self.workers_process.activity_state = ActivityState.Error
        self.workers_process.save()

        with self.assertNumQueries(16):
            response = self.client.post(reverse('api:process-retry', kwargs={'pk': self.workers_process.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        self.workers_process.refresh_from_db()
        self.assertEqual(self.workers_process.state, State.Unscheduled)
        self.assertEqual(self.workers_process.activity_state, ActivityState.Pending)
        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(self.workers_process))

    def test_retry_files(self):
        self.client.force_login(self.user)
        process = self.corpus.processes.create(mode=ProcessMode.Files, creator=self.user)
        process.worker_runs.create(version=self.version_with_model)
        process.tasks.create(state=State.Error, run=0, depth=0)
        self.assertEqual(process.state, State.Error)
        process.finished = timezone.now()

        with (
            self.settings(IMPORTS_WORKER_VERSION=str(self.version_with_model.id)),
            self.assertNumQueries(18),
        ):
            response = self.client.post(reverse('api:process-retry', kwargs={'pk': process.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        process.refresh_from_db()
        self.assertEqual(process.state, State.Unscheduled)
        self.assertIsNone(process.finished)
        self.assertQuerysetEqual(
            process.tasks.filter(run=1).values_list('slug', flat=True),
            ['import_files'],
        )

    def test_retry_s3(self):
        self.client.force_login(self.user)
        process = self.corpus.processes.create(
            mode=ProcessMode.S3,
            bucket_name='test',
            folder_type=self.volume_type,
            element_type=self.page_type,
            creator=self.user,
        )
        process.worker_runs.create(version=self.version_with_model)
        process.tasks.create(state=State.Error, run=0, depth=0)
        self.assertEqual(process.state, State.Error)
        process.finished = timezone.now()

        with (
            self.settings(
                IMPORTS_WORKER_VERSION=str(self.version_with_model.id),
                INGEST_IMAGESERVER_ID=str(ImageServer.objects.get(url='http://server').id),
            ),
            self.assertNumQueries(18),
        ):
            response = self.client.post(reverse('api:process-retry', kwargs={'pk': process.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        process.refresh_from_db()
        self.assertEqual(process.state, State.Unscheduled)
        self.assertIsNone(process.finished)
        self.assertQuerysetEqual(
            process.tasks.filter(run=1).values_list('slug', flat=True),
            ['import_s3'],
        )

    def test_retry_iiif(self):
        self.client.force_login(self.user)
        process = self.corpus.processes.create(mode=ProcessMode.IIIF, creator=self.user)
        process.worker_runs.create(version=self.version_with_model)
        process.tasks.create(state=State.Error, run=0, depth=0)
        self.assertEqual(process.state, State.Error)
        process.finished = timezone.now()

        with (
            self.settings(IMPORTS_WORKER_VERSION=str(self.version_with_model.id)),
            self.assertNumQueries(18),
        ):
            response = self.client.post(reverse('api:process-retry', kwargs={'pk': process.id}))
            self.assertEqual(response.status_code, status.HTTP_200_OK)

        process.refresh_from_db()
        self.assertEqual(process.state, State.Unscheduled)
        self.assertIsNone(process.finished)
        self.assertQuerysetEqual(
            process.tasks.filter(run=1).values_list('slug', flat=True),
            ['import_iiif'],
        )

    def test_from_files_requires_login(self):
        response = self.client.post(reverse('api:files-process'), {
            'files': [str(self.img_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_from_files(self):
        self.client.force_login(self.user)
        self.assertEqual(self.version_with_model.worker_runs.count(), 0)
        with self.settings(IMPORTS_WORKER_VERSION=str(self.version_with_model.id)):
            response = self.client.post(reverse('api:files-process'), {
                'files': [str(self.img_df.id)],
                'folder_type': 'volume',
                'element_type': 'page',
            }, format='json')
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data['id'])
        self.assertEqual(process.mode, ProcessMode.Files)
        self.assertListEqual(list(process.files.all()), [self.img_df])
        self.assertEqual(process.folder_type.slug, 'volume')
        self.assertEqual(process.element_type.slug, 'page')
        self.assertEqual(self.version_with_model.worker_runs.count(), 1)
        import_task = process.tasks.get(slug="import_files")
        self.assertEqual(
            import_task.env["ARKINDEX_WORKER_RUN_ID"],
            str(self.version_with_model.worker_runs.get().id),
        )

    @override_settings(IMPORTS_WORKER_VERSION=None)
    def test_from_files_pdf(self):
        self.client.force_login(self.user)
        with self.settings(IMPORTS_WORKER_VERSION=str(self.import_worker_version.id)):
            response = self.client.post(reverse('api:files-process'), {
                'files': [str(self.pdf_df.id)],
                'mode': 'files',
                'folder_type': 'volume',
                'element_type': 'page',
            }, format='json')
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data['id'])
        self.assertEqual(process.mode, ProcessMode.Files)
        self.assertListEqual(list(process.files.all()), [self.pdf_df])
        self.assertIsNone(process.element)

    @override_settings(IMPORTS_WORKER_VERSION=None)
    def test_from_files_image_and_pdf(self):
        self.client.force_login(self.user)
        with self.settings(IMPORTS_WORKER_VERSION=str(self.import_worker_version.id)):
            response = self.client.post(reverse('api:files-process'), {
                'files': [str(self.pdf_df.id), str(self.img_df.id)],
                'mode': 'files',
                'folder_type': 'volume',
                'element_type': 'page',
            }, format='json')
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data['id'])
        self.assertEqual(process.mode, ProcessMode.Files)
        self.assertListEqual(list(process.files.all()), [self.img_df, self.pdf_df])
        self.assertIsNone(process.element)

    @override_settings(
        PONOS_DEFAULT_ENV={'ARKINDEX_API_TOKEN': 'testToken'},
        ARKINDEX_TASKS_IMAGE='registry.teklia.com/tasks',
    )
    def test_from_files_iiif(self):
        self.client.force_login(self.user)

        with self.assertNumQueries(28), self.settings(IMPORTS_WORKER_VERSION=str(self.import_worker_version.id)):
            response = self.client.post(reverse('api:files-process'), {
                'files': [str(self.iiif_df.id)],
                'mode': 'iiif',
                'folder_type': 'volume',
                'element_type': 'page',
            }, format='json')
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        process = Process.objects.get(id=data['id'])

        self.assertEqual(process.mode, ProcessMode.IIIF)
        self.assertListEqual(list(process.files.all()), [self.iiif_df])
        self.assertEqual(process.folder_type.slug, 'volume')
        self.assertEqual(process.element_type.slug, 'page')
        self.assertIsNone(process.element)
        self.assertEqual(process.creator, self.user)
        self.assertEqual(process.corpus, self.corpus)
        self.assertIsNone(process.revision)

        worker_run = process.worker_runs.get()
        self.assertEqual(worker_run.version, self.import_worker_version)
        self.assertListEqual(worker_run.parents, [])
        self.assertIsNone(worker_run.configuration_id)
        self.assertIsNone(worker_run.model_version_id)

        self.assertTrue(process.tasks.exists())

    @override_settings(IMPORTS_WORKER_VERSION=None)
    def test_from_files_iiif_with_json_charset(self):
        self.iiif_df.content_type = 'application/json;charset=utf-8'
        self.assertEqual(self.iiif_df.content_type, 'application/json;charset=utf-8')

        self.client.force_login(self.user)
        with(self.settings(IMPORTS_WORKER_VERSION=str(self.import_worker_version.id))):
            response = self.client.post(reverse('api:files-process'), {
                'files': [str(self.iiif_df.id)],
                'mode': 'iiif',
                'folder_type': 'volume',
                'element_type': 'page',
            }, format='json')
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data['id'])
        self.assertEqual(process.mode, ProcessMode.IIIF)
        self.assertListEqual(list(process.files.all()), [self.iiif_df])
        self.assertEqual(process.folder_type.slug, 'volume')
        self.assertEqual(process.element_type.slug, 'page')
        self.assertIsNone(process.element)

    @override_settings(IMPORTS_WORKER_VERSION=None)
    def test_from_files_iiif_with_json_profile_and_charset(self):
        self.iiif_df.content_type = 'application/ld+json;profile="http://iiif.io/api/presentation/2/context.json";charset=utf-8'
        self.assertEqual(self.iiif_df.content_type, 'application/ld+json;profile="http://iiif.io/api/presentation/2/context.json";charset=utf-8')

        self.client.force_login(self.user)
        with(self.settings(IMPORTS_WORKER_VERSION=str(self.import_worker_version.id))):
            response = self.client.post(reverse('api:files-process'), {
                'files': [str(self.iiif_df.id)],
                'mode': 'iiif',
                'folder_type': 'volume',
                'element_type': 'page',
            }, format='json')
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data['id'])
        self.assertEqual(process.mode, ProcessMode.IIIF)
        self.assertListEqual(list(process.files.all()), [self.iiif_df])
        self.assertEqual(process.folder_type.slug, 'volume')
        self.assertEqual(process.element_type.slug, 'page')
        self.assertIsNone(process.element)

    def test_from_files_iiif_requires_folder_type(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:files-process'), {
            'files': [str(self.iiif_df.id)],
            'mode': 'iiif',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': ['IIIF imports require both folder_type and element_type']
        })

    def test_from_files_invalid_mode(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:files-process'), {
            'files': [str(self.img_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
            'mode': 'repository',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_from_files_iiif_wrong_type(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:files-process'), {
            'files': [str(self.pdf_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
            'mode': 'iiif',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'non_field_errors': ['IIIF imports can only import IIIF documents']})

    def test_from_files_files_wrong_type(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:files-process'), {
            'files': [str(self.iiif_df.id)],
            'folder_type': 'volume',
            'element_type': 'page',
            'mode': 'files',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'non_field_errors': ['File imports can only import images or PDF documents']})

    @override_settings(IMPORTS_WORKER_VERSION=None)
    def test_from_files_folder_id(self):
        self.client.force_login(self.user)
        volume = self.corpus.elements.get(name='Volume 1')
        with self.settings(IMPORTS_WORKER_VERSION=str(self.import_worker_version.id)):
            response = self.client.post(
                reverse('api:files-process'),
                {
                    'files': [str(self.pdf_df.id)],
                    'mode': 'files',
                    'folder_id': str(volume.id),
                    'element_type': 'page',
                },
                format='json',
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        process = Process.objects.get(id=data['id'])
        self.assertEqual(process.element, volume)

    def test_from_files_no_files(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:files-process'), {
            'files': [],
            'folder_type': 'volume',
            'element_type': 'page',
        }, format='json')
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_start_process_requires_login(self):
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(self.user_img_process.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_start_process_invalid_process_id(self):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': '12341234-1234-1234-1234-123412341234'})
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
        self.assertEqual(response.json(), {'detail': 'Not found.'})

    def test_start_process_wrong_process_mode(self):
        self.assertFalse(self.user_img_process.mode in (ProcessMode.Workers, ProcessMode.Dataset))
        # grant an admin access to this process
        self.user_img_process.corpus.memberships.create(user=self.user, level=Role.Admin.value)
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(self.user_img_process.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {'__all__': ['Only a Process with Workers or Dataset mode and not already launched can be started later on']}
        )

    def test_start_process_process_already_started(self):
        process2 = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        process2.run()
        self.assertTrue(process2.tasks.exists())

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(process2.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {'__all__': ['Only a Process with Workers or Dataset mode and not already launched can be started later on']}
        )

    def test_start_process_without_required_model(self):
        process2 = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        process2.worker_runs.create(version=self.version_with_model, parents=[], configuration=None)
        process2.save()

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(process2.id)})
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {"model_version": [f"The following workers require a model version and none was set: {[self.version_with_model.worker.name]}"]}
        )

    def test_start_process_with_required_model(self):
        process2 = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        run = process2.worker_runs.create(version=self.version_with_model, parents=[], configuration=None)
        run.model_version = self.model_version_1
        run.save()
        self.assertFalse(process2.tasks.exists())

        self.client.force_login(self.user)
        with self.assertNumQueries(19):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process2.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(response.json()['id'], str(process2.id))

    def test_start_process_empty(self):
        process2 = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)

        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process2.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(
            response.json(),
            {'__all__': ['The process must either use thumbnail generation or have worker runs.']},
        )

    def test_start_process_unavailable_worker_version(self):
        process2 = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        process2.worker_runs.create(version=self.recognizer, parents=[], configuration=None)
        self.recognizer.state = WorkerVersionState.Error
        self.recognizer.save()
        self.assertFalse(process2.tasks.exists())

        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process2.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {'version': ['This process contains one or more unavailable worker versions and cannot be started.']},
        )

    def test_start_process_unavailable_model_version(self):
        process2 = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        process2.worker_runs.create(version=self.recognizer, parents=[], configuration=None, model_version=self.model_version_1)
        self.model_version_1.state = ModelVersionState.Error
        self.model_version_1.save()
        self.assertFalse(process2.tasks.exists())

        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process2.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(
            response.json(),
            {'model_version': ['This process contains one or more unavailable model versions and cannot be started.']},
        )

    def test_start_process_required_fields_no_config(self):
        # Both workers now have a required field without a default value
        self.dla.configuration['user_configuration'] = {
            'some_field': {
                'title': 'pls configure me',
                'type': 'bool',
                'required': True,
            }
        }
        self.dla.save()
        self.recognizer.configuration['user_configuration'] = {
            'some_field': {
                'title': 'pls configure me',
                'type': 'bool',
                'required': True,
            }
        }
        self.recognizer.save()

        process2 = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        # One worker version is used without any configuration
        process2.worker_runs.create(
            version=self.recognizer,
            configuration=None,
            model_version=None,
        )
        # The other version is used with a configuration missing the required field
        process2.worker_runs.create(
            version=self.dla,
            configuration=self.dla.worker.configurations.create(
                name='oh no',
                configuration={
                    'not_some_field': 'oops',
                },
            ),
            model_version=None,
        )
        self.client.force_login(self.user)

        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process2.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {
            'worker_configuration': [
                'The following workers have required configuration fields that have not been set: '
                'Document layout analyser, Recognizer',
            ],
        })

    def test_start_process_workers(self):
        """
        A user can start a process with no parameters.
        Default chunks, thumbnails and farm are used. Nor cache or workers activity is set.
        """
        process2 = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        run = process2.worker_runs.create(version=self.recognizer, parents=[], configuration=None)
        self.assertFalse(process2.tasks.exists())

        self.client.force_login(self.user)

        with self.assertNumQueries(19):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process2.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(response.json()['id'], str(process2.id))
        process2.refresh_from_db()
        self.assertEqual(process2.state, State.Unscheduled)
        # Ensure default parameters are used
        self.assertEqual(process2.farm_id, get_default_farm_id())
        self.assertEqual(process2.tasks.count(), 2)
        task1, task2 = process2.tasks.order_by('slug')
        self.assertEqual(task1.slug, 'initialisation')
        self.assertEqual(task2.slug, f'reco_{str(run.id)[:6]}')
        self.assertIn('--chunks-number 1', task1.command)

    def test_start_process_dataset_requires_datasets(self):
        process2 = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Dataset)
        process2.worker_runs.create(version=self.recognizer, parents=[], configuration=None)
        self.assertFalse(process2.tasks.exists())

        self.client.force_login(self.user)

        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process2.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': ['A dataset process cannot be started if it does not have any associated datasets.']
        })

    def test_start_process_dataset_requires_dataset_in_same_corpus(self):
        process2 = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Dataset)
        process2.datasets.set([self.private_dataset])
        process2.worker_runs.create(version=self.recognizer, parents=[], configuration=None)
        self.assertFalse(process2.tasks.exists())

        self.client.force_login(self.user)

        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process2.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': ['At least one of the process datasets must be from the same corpus as the process.']
        })

    def test_start_process_dataset_unsupported_parameters(self):
        process2 = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Dataset)
        process2.datasets.set([self.dataset1, self.private_dataset])
        process2.worker_runs.create(version=self.recognizer, parents=[], configuration=None)

        self.client.force_login(self.user)

        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process2.id)}),
                {
                    'thumbnails': True,
                    'use_cache': True,
                    'worker_activity': True,
                }
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            'thumbnails': ['Thumbnails generation is not supported on Dataset processes.'],
            'use_cache': ['Caching is not supported on Dataset processes.'],
            'worker_activity': ['Worker activities are not supported on Dataset processes.'],
        })

    def test_start_process_dataset(self):
        process2 = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Dataset)
        process2.datasets.set([self.dataset1, self.private_dataset])
        run = process2.worker_runs.create(version=self.recognizer, parents=[], configuration=None)
        self.assertFalse(process2.tasks.exists())

        self.client.force_login(self.user)

        with self.assertNumQueries(17):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process2.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(response.json()['id'], str(process2.id))
        process2.refresh_from_db()
        self.assertEqual(process2.state, State.Unscheduled)
        # Ensure default parameters are used
        self.assertEqual(process2.farm_id, get_default_farm_id())
        self.assertEqual(process2.tasks.count(), 1)
        task = process2.tasks.get()
        self.assertEqual(task.slug, run.task_slug)
        self.assertQuerysetEqual(process2.datasets.order_by('name'), [
            self.private_dataset, self.dataset1
        ])

    def test_start_process_from_docker_image(self):
        """
        Start a process with an available WorkerVersion that only has a docker_image_iid
        """
        self.recognizer.docker_image_id = None
        self.recognizer.docker_image_iid = 'hello-world'
        self.recognizer.save()
        self.assertEqual(self.recognizer.state, WorkerVersionState.Available)

        process = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        run = process.worker_runs.create(version=self.recognizer, parents=[], configuration=None)
        self.assertFalse(process.tasks.exists())

        self.client.force_login(self.user)
        with self.assertNumQueries(18):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(response.json()['id'], str(process.id))
        process.refresh_from_db()
        self.assertEqual(process.state, State.Unscheduled)
        task1, task2 = process.tasks.order_by('slug')
        self.assertEqual(task1.slug, 'initialisation')
        self.assertEqual(task2.slug, run.task_slug)

    def test_start_process_select_farm_id(self):
        """
        A user can specify a ponos farm to use for a process
        """
        # Run get_default_farm_id once to ensure the new farm is not the default one
        get_default_farm_id()
        barley_farm = Farm.objects.create(name='Barley farm')
        self.assertNotEqual(get_default_farm_id(), barley_farm.id)
        workers_process = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        workers_process.worker_runs.create(version=self.recognizer, parents=[], configuration=None)
        self.client.force_login(self.user)

        with self.assertNumQueries(19):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(workers_process.id)}),
                {'farm': str(barley_farm.id)}
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        workers_process.refresh_from_db()
        self.assertEqual(workers_process.state, State.Unscheduled)
        self.assertEqual(workers_process.farm_id, barley_farm.id)

    def test_start_process_wrong_farm_id(self):
        """
        Handle a payload with a wrong farm ID starting a process
        """
        workers_process = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        self.client.force_login(self.user)
        wrong_farm_id = uuid.uuid4()

        with self.assertNumQueries(9):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(workers_process.id)}),
                {'farm': str(wrong_farm_id)}
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertEqual(response.json(), {'farm': [f'Invalid pk "{wrong_farm_id}" - object does not exist.']})

    def test_start_process_params_validation(self):
        process = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        self.client.force_login(self.user)
        wrong_params_checks = [
            ({'chunks': 0}, {'chunks': ['Ensure this value is greater than or equal to 1.']}),
            ({'chunks': 20}, {'chunks': ['Ensure this value is less than or equal to 10.']}),
            ({'chunks': 'max'}, {'chunks': ['A valid integer is required.']}),
            ({'thumbnails': 'gloubiboulga'}, {'thumbnails': ['Must be a valid boolean.']})
        ]
        for (params, check) in wrong_params_checks:
            with self.subTest(**params), self.assertNumQueries(8):
                response = self.client.post(reverse('api:process-start', kwargs={'pk': str(process.id)}), params)
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(response.json(), check)

    @override_settings(MAX_CHUNKS=42)
    def test_start_process_serializer_chunks_config(self):
        """
        StartProcess should restrict the chunks to `settings.MAX_CHUNKS`
        """
        process = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        process.worker_runs.create(version=self.recognizer, parents=[], configuration=None)
        self.client.force_login(self.user)

        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process.id)}),
                {'chunks': 43},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            'chunks': ['Ensure this value is less than or equal to 42.'],
        })

    def test_start_process_workers_parameters(self):
        """
        It should be possible to pass chunks and thumbnails parameters when starting a workers process
        """
        process = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Workers)
        # Add a worker run to this process
        run = process.worker_runs.create(version=self.recognizer, parents=[], configuration=None)

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(process.id)}),
            {'chunks': 3, 'thumbnails': 'true'}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        process.refresh_from_db()

        self.assertEqual(list(process.tasks.order_by('slug').values_list('slug', flat=True)), [
            'initialisation',
            f'{run.task_slug}_1',
            f'{run.task_slug}_2',
            f'{run.task_slug}_3',
            'thumbnails_1',
            'thumbnails_2',
            'thumbnails_3'
        ])

    def test_start_process_dataset_chunks(self):
        """
        It should be possible to pass chunks when starting a dataset process
        """
        process = self.corpus.processes.create(creator=self.user, mode=ProcessMode.Dataset)
        process.datasets.set([self.dataset1, self.dataset2])
        # Add a worker run to this process
        run = process.worker_runs.create(version=self.recognizer, parents=[], configuration=None)

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:process-start', kwargs={'pk': str(process.id)}),
            {'chunks': 3}
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        process.refresh_from_db()

        self.assertEqual(list(process.tasks.order_by('slug').values_list('slug', flat=True)), [
            f'{run.task_slug}_1',
            f'{run.task_slug}_2',
            f'{run.task_slug}_3',
        ])

    @patch('arkindex.process.models.Process.worker_runs')
    @patch('arkindex.project.triggers.process_tasks.initialize_activity.delay')
    def test_start_process_options_requires_workers(self, activities_delay_mock, worker_runs_mock):
        """
        Cache and worker activity options can be triggered when starting a process
        """
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            element_type=self.corpus.types.get(slug='page')
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process.id)}),
                {'use_cache': 'true', 'worker_activity': 'true', 'use_gpu': 'true'}
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            '__all__': ['The process must either use thumbnail generation or have worker runs.'],
            'use_cache': ['The process must have workers attached to use cached results.'],
            'worker_activity': ['The process must have workers attached to handle their activity.'],
            'use_gpu': ['The process must have workers attached to use GPUs.'],
        })
        process.refresh_from_db()
        self.assertFalse(process.use_cache)
        self.assertEqual(process.activity_state, ActivityState.Disabled)

    @patch('arkindex.process.models.Process.worker_runs')
    @patch('arkindex.process.models.Process.versions')
    @patch('arkindex.project.triggers.process_tasks.initialize_activity.delay')
    def test_start_process_options(self, activities_delay_mock, versions_mock, worker_runs_mock):
        """
        Cache and worker activity options requires the process to have workers attached
        """
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            element_type=self.corpus.types.get(slug='page')
        )
        self.assertFalse(process.use_cache)
        self.assertEqual(process.activity_state, ActivityState.Disabled)

        # Add a worker run to this process
        run_mock = MagicMock()
        run_mock.version = self.version_gpu
        run_mock.version_id = self.version_gpu.id
        run_mock.model_version_id = None
        run_mock.configuration_id = None
        task = Task(run=0, slug='task_slug', priority=1, process=process)
        run_mock.build_task.return_value = task, []

        # Cheat to mock a query on process worker runs
        worker_runs_mock.all.return_value = [run_mock]

        self.client.force_login(self.user)

        with self.assertNumQueries(17):
            response = self.client.post(
                reverse('api:process-start', kwargs={'pk': str(process.id)}),
                {'use_cache': 'true', 'worker_activity': 'true', 'use_gpu': 'true'}
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        process.refresh_from_db()
        self.assertTrue(process.use_cache)
        self.assertTrue(process.use_gpu)
        # The process activity state is set to Pending while activities are created in an async job
        self.assertEqual(process.activity_state, ActivityState.Pending)
        self.assertEqual(activities_delay_mock.call_count, 1)

    def test_start_duplicated_worker_version(self):
        """
        A WorkerRun depending on another WorkerRun of the same WorkerVersion should be supported
        """
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
        )
        run_1 = process.worker_runs.create(
            version=self.recognizer,
            configuration=self.recognizer.worker.configurations.create(
                name='some_config',
                configuration={"a": "b"},
            ),
        )
        run_2 = process.worker_runs.create(
            version=self.recognizer,
            parents=[run_1.id],
        )
        self.assertNotEqual(run_1.task_slug, run_2.task_slug)

        self.client.force_login(self.user)
        with self.assertNumQueries(20):
            response = self.client.post(reverse('api:process-start', kwargs={'pk': str(process.id)}))
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(process.tasks.count(), 3)
        init_task = process.tasks.get(slug='initialisation')
        task_1 = process.tasks.get(slug=run_1.task_slug)
        task_2 = process.tasks.get(slug=run_2.task_slug)

        self.assertQuerysetEqual(task_1.parents.all(), [init_task])
        self.assertQuerysetEqual(task_2.parents.all(), [task_1])

    @override_settings(PONOS_DEFAULT_ENV={'ARKINDEX_API_TOKEN': 'testToken'})
    @override_settings(ARKINDEX_TASKS_IMAGE='registry.teklia.com/tasks')
    @patch('arkindex.process.builder.task_token_default')
    def test_pdf_import_run(self, token_mock):
        """
        Build tasks for a PDF import with a worker version defined in settings
        """
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Files,
        )
        process.files.create(name='A PDF file', size=10e6, content_type='pdf', corpus=self.corpus)
        token_mock.return_value = '12345'

        with self.settings(IMPORTS_WORKER_VERSION=str(self.import_worker_version.id)):
            process.run()

        worker_run = process.worker_runs.get()
        self.assertEqual(worker_run.version, self.import_worker_version)
        self.assertListEqual(worker_run.parents, [])
        self.assertIsNone(worker_run.configuration_id)
        self.assertIsNone(worker_run.model_version_id)

        import_task = process.tasks.get(slug='import_files')
        self.assertEqual(import_task.env, {
            'ARKINDEX_API_TOKEN': 'testToken',
            'ARKINDEX_PROCESS_ID': str(process.id),
            'ARKINDEX_CORPUS_ID': str(self.corpus.id),
            'ARKINDEX_WORKER_RUN_ID': str(worker_run.id),
            'ARKINDEX_TASK_TOKEN': '12345'
        })
        self.assertEqual(import_task.command, f'python -m arkindex_tasks.import_files {process.id}')
        self.assertEqual(import_task.image, 'registry.teklia.com/tasks')

    @override_settings(PONOS_DEFAULT_ENV={'ARKINDEX_API_TOKEN': 'testToken'})
    @override_settings(ARKINDEX_TASKS_IMAGE='registry.teklia.com/tasks')
    @override_settings(PUBLIC_HOSTNAME='https://arkindex.localhost')
    @patch('arkindex.ponos.models.base64.encodebytes')
    def test_worker_run_model_version_run(self, token_mock):
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
        )
        token_mock.side_effect = [b'12345', b'78945']
        run = process.worker_runs.create(version=self.version_with_model, parents=[], configuration=None)
        run.model_version = self.model_version_1
        run.save()
        process.run()

        initialisation_task = process.tasks.get(slug='initialisation')
        self.assertEqual(initialisation_task.env, {
            'ARKINDEX_API_TOKEN': 'testToken',
            'ARKINDEX_PROCESS_ID': str(process.id),
            'ARKINDEX_CORPUS_ID': str(self.corpus.id),
            'ARKINDEX_TASK_TOKEN': '12345'
        })
        self.assertEqual(initialisation_task.command, f'python -m arkindex_tasks.init_elements {process.id} --chunks-number 1')
        self.assertEqual(initialisation_task.image, 'registry.teklia.com/tasks')

        worker_task = process.tasks.get(slug=run.task_slug)
        self.assertEqual(worker_task.env, {
            'ARKINDEX_API_TOKEN': 'testToken',
            'ARKINDEX_PROCESS_ID': str(process.id),
            'ARKINDEX_CORPUS_ID': str(self.corpus.id),
            'ARKINDEX_TASK_TOKEN': '78945',
            'ARKINDEX_WORKER_RUN_ID': str(run.id),
            'TASK_ELEMENTS': '/data/initialisation/elements.json'
        })
        self.assertEqual(worker_task.image, run.version.docker_image_iid or run.version.docker_image_name)
        self.assertEqual(worker_task.command, run.version.docker_command)
        self.assertEqual(worker_task.image_artifact.id, run.version.docker_image_id)
        self.assertEqual(worker_task.extra_files, {
            "model": "https://arkindex.localhost" + reverse('api:model-version-download', kwargs={'pk': run.model_version.id}) + f"?token={run.model_version.build_authentication_token_hash()}"
        })
        self.assertEqual(len(worker_task.parents.all()), 1)
        self.assertEqual(worker_task.parents.first().slug, 'initialisation')

    def test_clear_process(self):
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            element_type=self.corpus.types.get(slug='page')
        )
        process.worker_runs.create(version=self.dla, parents=[], configuration=None)
        process.worker_runs.create(version=self.recognizer, parents=[], configuration=None)
        self.assertEqual(process.worker_runs.count(), 2)

        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.delete(reverse('api:clear-process', kwargs={'pk': str(process.id)}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        process.refresh_from_db()
        self.assertEqual(process.worker_runs.count(), 0)

    def test_clear_process_requires_login(self):
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            element_type=self.corpus.types.get(slug='page')
        )
        process.worker_runs.create(version=self.dla, parents=[], configuration=None)
        process.worker_runs.create(version=self.recognizer, parents=[], configuration=None)
        self.assertEqual(process.worker_runs.count(), 2)

        with self.assertNumQueries(0):
            response = self.client.delete(reverse('api:clear-process', kwargs={'pk': str(process.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_clear_process_requires_verified(self):
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            element_type=self.corpus.types.get(slug='page')
        )
        process.worker_runs.create(version=self.dla, parents=[], configuration=None)
        process.worker_runs.create(version=self.recognizer, parents=[], configuration=None)
        self.assertEqual(process.worker_runs.count(), 2)

        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.delete(reverse('api:clear-process', kwargs={'pk': str(process.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have permission to perform this action.'})

    def test_clear_process_does_not_exist(self):
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            element_type=self.corpus.types.get(slug='page')
        )
        process.worker_runs.create(version=self.dla, parents=[], configuration=None)
        process.worker_runs.create(version=self.recognizer, parents=[], configuration=None)
        self.assertEqual(process.worker_runs.count(), 2)

        fake = uuid.uuid4()

        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.delete(reverse('api:clear-process', kwargs={'pk': str(fake)}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_clear_process_not_unscheduled(self):
        """
        Cannot clear a process that has already started
        """
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            element_type=self.corpus.types.get(slug='page')
        )
        process.worker_runs.create(version=self.dla, parents=[], configuration=None)
        process.worker_runs.create(version=self.recognizer, parents=[], configuration=None)
        self.assertEqual(process.worker_runs.count(), 2)
        process.run()
        process.tasks.update(state=State.Running)
        self.assertEqual(process.state, State.Running)

        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.delete(reverse('api:clear-process', kwargs={'pk': str(process.id)}))
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'__all__': ['A process can only be cleared before getting started.']})

    def test_clear_process_unscheduled(self):
        """
        Cannot clear a process that has tasks, even unscheduled
        """
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            element_type=self.corpus.types.get(slug='page')
        )
        process.worker_runs.create(version=self.dla, parents=[], configuration=None)
        process.worker_runs.create(version=self.recognizer, parents=[], configuration=None)
        self.assertEqual(process.worker_runs.count(), 2)
        process.run()
        self.assertEqual(process.state, State.Unscheduled)

        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.delete(reverse('api:clear-process', kwargs={'pk': str(process.id)}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'__all__': ['A process can only be cleared before getting started.']})

    def test_clear_process_requires_permissions(self):
        process = self.corpus.processes.create(
            creator=self.user,
            mode=ProcessMode.Workers,
            element_type=self.corpus.types.get(slug='page')
        )
        process.worker_runs.create(version=self.dla, parents=[], configuration=None)
        process.worker_runs.create(version=self.recognizer, parents=[], configuration=None)
        self.assertEqual(process.worker_runs.count(), 2)

        user2 = User.objects.create_user('email@mail.com', 'bob')
        user2.verified_email = True
        user2.save()

        self.corpus.memberships.create(user=user2, level=Role.Contributor.value)
        self.client.force_login(user2)
        with self.assertNumQueries(6):
            response = self.client.delete(reverse('api:clear-process', kwargs={'pk': str(process.id)}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have a sufficient access level to this process.'})

    def test_select_failed_elts_forbidden_methods(self):
        self.client.force_login(self.user)
        methods = (self.client.get, self.client.put, self.client.delete)
        for method in methods:
            response = self.client.delete(reverse('api:process-select-failures', kwargs={'pk': str(self.elts_process.id)}))
            self.assertEqual(response.status_code, status.HTTP_405_METHOD_NOT_ALLOWED)

    def test_select_failed_elts_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.delete(reverse('api:process-select-failures', kwargs={'pk': str(self.elts_process.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_select_failed_elts_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(reverse('api:process-select-failures', kwargs={'pk': str(self.elts_process.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_select_failed_elts_requires_valid_id(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(3):
            response = self.client.post(reverse('api:process-select-failures', kwargs={'pk': str(uuid.uuid4())}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_select_failed_elts_requires_corpus_read_access(self):
        self.client.force_login(self.user)
        self.elts_process.corpus.memberships.filter(user=self.user).delete()
        self.elts_process.corpus.public = False
        self.elts_process.corpus.save()
        with self.assertNumQueries(5):
            response = self.client.post(reverse('api:process-select-failures', kwargs={'pk': str(self.elts_process.id)}))
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have a read access to the corpus of this process.'})

    def test_select_failed_elts_requires_workers_mode(self):
        self.client.force_login(self.user)
        self.user_img_process.corpus.memberships.create(user=self.user, level=Role.Contributor.value)
        with self.assertNumQueries(3):
            response = self.client.post(reverse('api:process-select-failures', kwargs={'pk': str(self.user_img_process.id)}))
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_select_failed_elts_requires_worker_activity(self):
        self.client.force_login(self.user)
        self.elts_process.activity_state = ActivityState.Disabled
        self.elts_process.save()
        with self.assertNumQueries(6):
            response = self.client.post(reverse('api:process-select-failures', kwargs={'pk': str(self.elts_process.id)}))
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'__all__': ['The process must use worker activities to select elements with failures.']},
        )

    def test_select_failed_elts_requires_finished_state(self):
        unfinished_states = {
            State.Running,
            State.Stopping,
            State.Pending,
            State.Unscheduled,
        }
        self.elts_process.run()
        self.assertEqual(self.elts_process.tasks.get().run, 0)
        self.elts_process.activity_state = ActivityState.Ready
        self.elts_process.save()
        self.client.force_login(self.user)
        for state in unfinished_states:
            with self.subTest(state=state):
                self.elts_process.tasks.update(state=state)
                self.assertEqual(self.elts_process.state, state)
                response = self.client.post(
                    reverse('api:process-select-failures', kwargs={'pk': str(self.elts_process.id)})
                )
                self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertDictEqual(
                response.json(),
                {'__all__': ['The process must be finished to select elements with failures.']},
            )

    def test_select_failed_elts_no_failure(self):
        self.elts_process.run()
        self.assertEqual(self.elts_process.tasks.get().run, 0)
        self.elts_process.tasks.update(state=State.Completed)
        self.elts_process.activity_state = ActivityState.Ready
        self.elts_process.save()
        # Process has a single element that is ok
        elt = self.corpus.elements.get(name='Volume 1')
        self.elts_process.elements.add(elt.id)
        WorkerActivity.objects.create(
            process=self.elts_process,
            worker_version=self.recognizer,
            element=elt,
            state=WorkerActivityState.Processed,
        )
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse('api:process-select-failures', kwargs={'pk': str(self.elts_process.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {'__all__': ['This process has no element in a failure state.']},
        )

    def test_select_failed_elts(self):
        self.elts_process.run()
        self.assertEqual(self.elts_process.tasks.get().run, 0)
        self.elts_process.tasks.update(state=State.Completed)
        self.elts_process.activity_state = ActivityState.Ready
        self.elts_process.save()

        elements = self.corpus.elements.filter(
            name__in=(
                'Volume 1, page 1r',
                'Volume 1, page 1v',
                'Volume 1, page 2r',
            )
        )
        # Only recto pages has an activity with a failure
        for elt in elements:
            self.elts_process.elements.add(elt.id)
            WorkerActivity.objects.create(
                process=self.elts_process,
                worker_version=self.recognizer,
                element=elt,
                state=WorkerActivityState.Processed if elt.name.endswith("v") else WorkerActivityState.Error,
            )

        self.assertCountEqual(list(self.user.selections.values_list("element__name", flat=True)), [])
        self.client.force_login(self.user)
        with self.assertNumQueries(9):
            response = self.client.post(
                reverse('api:process-select-failures', kwargs={'pk': str(self.elts_process.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)

        self.assertCountEqual(
            list(self.user.selections.values_list("element__name", flat=True)),
            ['Volume 1, page 1r', 'Volume 1, page 2r'],
        )
